import functools
import datetime
from typing import List, Tuple, Dict, Set, Union

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList, SearchableLazyList, SearchableDictionary

from ..configuration import settings
from ..data import Region, Platform, Tier, Map, GameType, GameMode, Queue, Division, Side, Season
from .common import CoreData, DataObjectList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaGhostList
from ..dto import match as dto
from .summoner import Summoner
from .staticdata.champion import Champion
from .staticdata.rune import Rune
from .staticdata.mastery import Mastery
from .staticdata.summonerspell import SummonerSpell
from .staticdata.item import Item


##############
# Data Types #
##############


class MatchListData(DataObjectList):
    _dto_type = dto.MatchListDto
    _renamed = {"account_id": "accountId", "seasons": "season", "champion_ids": "champion", "queues": "queue", "being_index": "beginIndex", "end_index": "endIndex", "begin_time": "beginTime", "end_time": "endTime"}

    @property
    def account_id(self) -> str:
        return self._dto["accountId"]

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def queues(self) -> Set[int]:
        return self._dto["queue"]

    @property
    def seasons(self) -> Set[int]:
        return self._dto["season"]

    @property
    def champion_ids(self) -> Set[int]:
        return self._dto["champion"]

    @property
    def begin_index(self) -> int:
        return self._dto["beginIndex"]

    @property
    def end_index(self) -> int:
        return self._dto["endIndex"]

    @property
    def begin_time(self) -> int:
        return self._dto["beginTime"]

    @property
    def end_time(self) -> int:
        return self._dto["endTime"]


class PositionData(CoreData):
    _renamed = {}

    @property
    def x(self) -> int:
        return self._dto["x"]

    @property
    def y(self) -> int:
        return self._dto["y"]


class EventData(CoreData):
    _renamed = {"event_type": "eventType", "tower_type": "towerType", "team_id": "teamId", "ascended_type": "ascendedType", "killer_id": "killerId", "level_up_type": "levelUpType", "point_captured": "pointCaptured", "assisting_participant_ids": "assistingParticipantIds", "ward_type": "wardType", "monster_type": "monsterType", "skill_slot": "skillSlot", "victim_id": "victimId", "after_id": "afterId", "monster_sub_type": "monsterSubType", "lane_type": "laneType", "item_id": "itemId", "participant_id": "participantId", "building_type": "buildingType", "creator_id": "creatorId", "before_id": "beforeId"}

    @property
    def event_type(self) -> str:
        return self._dto["eventType"]

    @property
    def tower_type(self) -> str:
        return self._dto["towerType"]

    @property
    def team_id(self) -> int:
        return self._dto["teamId"]

    @property
    def ascended_type(self) -> str:
        return self._dto["ascendedType"]

    @property
    def killer_id(self) -> int:
        return self._dto["killerId"]

    @property
    def level_up_type(self) -> str:
        return self._dto["levelUpType"]

    @property
    def point_captured(self) -> str:
        return self._dto["pointCaptured"]

    @property
    def assisting_participant_ids(self) -> List[int]:
        return self._dto["assistingParticipantIds"]

    @property
    def ward_type(self) -> str:
        return self._dto["wardType"]

    @property
    def monster_type(self) -> str:
        return self._dto["monsterType"]

    @property
    def type(self) -> List[str]:
        """Legal values: CHAMPION_KILL, WARD_PLACED, WARD_KILL, BUILDING_KILL, ELITE_MONSTER_KILL, ITEM_PURCHASED, ITEM_SOLD, ITEM_DESTROYED, ITEM_UNDO, SKILL_LEVEL_UP, ASCENDED_EVENT, CAPTURE_POINT, PORO_KING_SUMMON"""
        return self._dto["type"]

    @property
    def skill_slot(self) -> int:
        return self._dto["skillSlot"]

    @property
    def victim_id(self) -> int:
        return self._dto["victimId"]

    @property
    def timestamp(self) -> int:
        return self._dto["timestamp"]

    @property
    def after_id(self) -> int:
        return self._dto["afterId"]

    @property
    def monster_sub_type(self) -> str:
        return self._dto["monsterSubType"]

    @property
    def lane_type(self) -> str:
        return self._dto["laneType"]

    @property
    def item_id(self) -> int:
        return self._dto["itemId"]

    @property
    def participant_id(self) -> int:
        return self._dto["participantId"]

    @property
    def building_type(self) -> str:
        return self._dto["buildingType"]

    @property
    def creator_id(self) -> int:
        return self._dto["creatorId"]

    @property
    def position(self) -> PositionData:
        return PositionData.from_dto(self._dto["position"])

    @property
    def before_id(self) -> int:
        return self._dto["beforeId"]


class ParticipantFrameData(CoreData):
    _renamed = {"total_gold": "totalGold", "team_score": "teamScore", "participant_id": "participantId", "current_gold": "currentGold", "minions_killed": "minionsKilled", "dominion_score": "dominionScore", "jungle_minions_killed": "jungleMinionsKilled"}

    @property
    def total_gold(self) -> int:
        return self._dto["totalGold"]

    @property
    def team_score(self) -> int:
        return self._dto["teamScore"]

    @property
    def participant_id(self) -> int:
        return self._dto["participantId"]

    @property
    def level(self) -> int:
        return self._dto["level"]

    @property
    def current_gold(self) -> int:
        return self._dto["currentGold"]

    @property
    def minions_killed(self) -> int:
        return self._dto["minionsKilled"]

    @property
    def dominion_score(self) -> int:
        return self._dto["dominionScore"]

    @property
    def position(self) -> PositionData:
        return PositionData.from_dto(self._dto["position"])

    @property
    def xp(self) -> int:
        return self._dto["xp"]

    @property
    def jungle_minions_killed(self) -> int:
        return self._dto["jungleMinionsKilled"]


class FrameData(CoreData):
    _renamed = {"participant_frames": "participantFrames"}

    @property
    def timestamp(self) -> int:
        return self._dto["timestamp"]

    @property
    def participant_frames(self) -> Dict[int, ParticipantFrameData]:
        return {k: ParticipantFrameData.from_dto(v) for k, v in self._dto["participantFrames"].items()}

    @property
    def events(self) -> List[EventData]:
        return [EventData.from_dto(event) for event in self._dto["events"]]


class TimelineData(CoreData):
    _dto_type = dto.TimelineDto
    _renamed = {"id": "matchId"}

    @property
    def id(self) -> int:
        return self._dto["matchId"]

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def frames(self) -> List[FrameData]:
        return [FrameData.from_dto(frame) for frame in self._dto["frames"]]

    @property
    def frame_interval(self) -> int:
        return self._dto["frameInterval"]


class ParticipantTimelineData(CoreData):
    _renamed = {"id": "participantId", "cs_diff_per_min_deltas": "csDiffPerMinDeltas", "gold_per_min_deltas": "goldPerMinDeltas", "xp_pdiff_per_min_deltas": "xpDiffPerMinDeltas", "creeps_per_min_deltas": "creepsPerMinDeltas", "damage_taken_diff_per_min_deltas": "damageTakenDiffPerMinDeltas", "damage_taken_per_min_deltas": "damageTakenPerMinDeltas"}

    @property
    def lane(self) -> str:
        return self._dto["lane"]

    @property
    def role(self) -> str:
        return self._dto["role"]

    @property
    def id(self) -> int:
        return self._dto["participantId"]

    @property
    def cs_diff_per_min_deltas(self) -> Dict[str, float]:
        return self._dto["csDiffPerMinDeltas"]

    @property
    def gold_per_min_deltas(self) -> Dict[str, float]:
        return self._dto["goldPerMinDeltas"]

    @property
    def xp_diff_per_min_deltas(self) -> Dict[str, float]:
        return self._dto["xpDiffPerMinDeltas"]

    @property
    def creeps_per_min_deltas(self) -> Dict[str, float]:
        return self._dto["creepsPerMinDeltas"]

    @property
    def xp_per_min_deltas(self) -> Dict[str, float]:
        return self._dto["xpPerMinDeltas"]

    @property
    def damage_taken_per_min_deltas(self) -> Dict[str, float]:
        return self._dto["damageTakenPerMinDeltas"]

    @property
    def damage_taken_diff_per_min_deltas(self) -> Dict[str, float]:
        return self._dto["damageTakenDiffPerMinDeltas"]


class ParticipantStatsData(CoreData):
    _renamed = {"physical_damage_dealt": "physicalDamageDealt", "magic_damage_dealt": "magicDamageDealt", "neutral_minions_killed_team_jungle": "neutralMinionsKilledTeamJungle", "total_player_score": "totalPlayerScore", "neutral_minions_killed_enemy_jungle": "neutralMinionsKilledEnemyJungle", "altars_captured": "altarsCaptured", "largest_critical_strike": "largestCriticalStrike", "total_damage_dealt": "totalDamageDealt", "magic_damage_dealt_to_champions": "magicDamageDealtToChampions", "vision_wards_bought_in_game": "visionWardsBoughtInGame", "damage_dealt_to_objectives": "damageDealtToObjectives", "largest_killing_spree": "largestKillingSpree", "quadra_kills": "quadraKills", "team_objective": "teamObjective", "total_time_crowd_control_dealt": "totalTimeCrowdControlDealt", "longest_time_spent_living": "longestTimeSpentLiving", "wards_killed": "wardsKilled", "first_tower_assist": "firstTowerAssist", "first_tower_kill": "firstTowerKill", "first_blood_assist": "firstBloodAssist", "vision_score": "visionScore", "wards_placed": "wardsPlaced", "turret_kills": "turretKills", "triple_kills": "tripleKills", "damage_self_mitigated": "damageSelfMitigated", "champion_level": "champLevel", "node_neutralize_assist": "nodeNeutralizeAssist", "first_inhibitor_kill": "firstInhibitorKill", "gold_earned": "goldEarned", "magical_damage_taken": "magicalDamageTaken", "double_kills": "doubleKills", "node_capture_assist": "nodeCaptureAssist", "true_damage_taken": "trueDamageTaken", "node_neutralize": "nodeNeutralize", "first_inhibitor_assist": "firstInhibitorAssist", "unreal_kills": "unrealKills", "neutral_minions_killed": "neutralMinionsKilled", "objective_player_score": "objectivePlayerScore", "combat_player_score": "combatPlayerScore", "damage_dealt_to_turrets": "damageDealtToTurrets", "altars_neutralized": "altarsNeutralized", "physical_damage_dealt_to_champions": "physicalDamageDealtToChampions", "gold_spent": "goldSpent", "true_damage_dealt": "trueDamageDealt", "true_damage_dealt_to_champions": "trueDamageDealtToChampions", "id": "participantId", "penta_kills": "pentaKills", "total_heal": "totalHeal", "total_minions_killed": "totalMinionsKilled", "first_blood_kill": "firstBloodKill", "node_capture": "nodeCapture", "largest_multi_kill": "largestMultiKill", "sight_wards_bought_in_game": "sightWardsBoughtInGame", "total_damage_dealt_to_champions": "totalDamageDealtToChampions", "total_units_healed": "totalUnitsHealed", "inhibitor_kills": "inhibitorKills", "total_score_rank": "totalScoreRank", "total_damage_taken": "totalDamageTaken", "killing_sprees": "killingSprees", "time_CCing_others": "timeCCingOthers", "physical_damage_taken": "physicalDamageTaken"}

    @property
    def physical_damage_dealt(self) -> int:
        return self._dto["physicalDamageDealt"]

    @property
    def magic_damage_dealt(self) -> int:
        return self._dto["magicDamageDealt"]

    @property
    def neutral_minions_killed_team_jungle(self) -> int:
        return self._dto["neutralMinionsKilledTeamJungle"]

    @property
    def total_player_score(self) -> int:
        return self._dto["totalPlayerScore"]

    @property
    def deaths(self) -> int:
        return self._dto["deaths"]

    @property
    def win(self) -> bool:
        return self._dto["win"]

    @property
    def neutral_minions_killed_enemy_jungle(self) -> int:
        return self._dto["neutralMinionsKilledEnemyJungle"]

    @property
    def altars_captured(self) -> int:
        return self._dto["altarsCaptured"]

    @property
    def largest_critical_strike(self) -> int:
        return self._dto["largestCriticalStrike"]

    @property
    def total_damage_dealt(self) -> int:
        return self._dto["totalDamageDealt"]

    @property
    def magic_damage_dealt_to_champions(self) -> int:
        return self._dto["magicDamageDealtToChampions"]

    @property
    def vision_wards_bought_in_game(self) -> int:
        return self._dto["visionWardsBoughtInGame"]

    @property
    def damage_dealt_to_objectives(self) -> int:
        return self._dto["damageDealtToObjectives"]

    @property
    def largest_killing_spree(self) -> int:
        return self._dto["largestKillingSpree"]

    @property
    def item1(self) -> int:
        return self._dto["item1"]

    @property
    def quadra_kills(self) -> int:
        return self._dto["quadraKills"]

    @property
    def team_objective(self) -> int:
        return self._dto["teamObjective"]

    @property
    def total_time_crowd_control_dealt(self) -> int:
        return self._dto["totalTimeCrowdControlDealt"]

    @property
    def longest_time_spent_living(self) -> int:
        return self._dto["longestTimeSpentLiving"]

    @property
    def wards_killed(self) -> int:
        return self._dto["wardsKilled"]

    @property
    def first_tower_assist(self) -> bool:
        return self._dto["firstTowerAssist"]

    @property
    def first_tower_kill(self) -> bool:
        return self._dto["firstTowerKill"]

    @property
    def item2(self) -> int:
        return self._dto["item2"]

    @property
    def item3(self) -> int:
        return self._dto["item3"]

    @property
    def item0(self) -> int:
        return self._dto["item0"]

    @property
    def first_blood_assist(self) -> bool:
        return self._dto["firstBloodAssist"]

    @property
    def vision_score(self) -> int:
        return self._dto["visionScore"]

    @property
    def wards_placed(self) -> int:
        return self._dto["wardsPlaced"]

    @property
    def item4(self) -> int:
        return self._dto["item4"]

    @property
    def item5(self) -> int:
        return self._dto["item5"]

    @property
    def item6(self) -> int:
        return self._dto["item6"]

    @property
    def turret_kills(self) -> int:
        return self._dto["turretKills"]

    @property
    def triple_kills(self) -> int:
        return self._dto["tripleKills"]

    @property
    def damage_self_mitigated(self) -> int:
        return self._dto["damageSelfMitigated"]

    @property
    def champion_level(self) -> int:
        return self._dto["champLevel"]

    @property
    def node_neutralize_assist(self) -> int:
        return self._dto["nodeNeutralizeAssist"]

    @property
    def first_inhibitor_kill(self) -> bool:
        return self._dto["firstInhibitorKill"]

    @property
    def gold_earned(self) -> int:
        return self._dto["goldEarned"]

    @property
    def magical_damage_taken(self) -> int:
        return self._dto["magicalDamageTaken"]

    @property
    def kills(self) -> int:
        return self._dto["kills"]

    @property
    def double_kills(self) -> int:
        return self._dto["doubleKills"]

    @property
    def node_capture_assist(self) -> int:
        return self._dto["nodeCaptureAssist"]

    @property
    def true_damage_taken(self) -> int:
        return self._dto["trueDamageTaken"]

    @property
    def node_neutralize(self) -> int:
        return self._dto["nodeNeutralize"]

    @property
    def first_inhibitor_assist(self) -> bool:
        return self._dto["firstInhibitorAssist"]

    @property
    def assists(self) -> int:
        return self._dto["assists"]

    @property
    def unreal_kills(self) -> int:
        return self._dto["unrealKills"]

    @property
    def neutral_minions_killed(self) -> int:
        return self._dto["neutralMinionsKilled"]

    @property
    def objective_player_score(self) -> int:
        return self._dto["objectivePlayerScore"]

    @property
    def combat_player_score(self) -> int:
        return self._dto["combatPlayerScore"]

    @property
    def damage_dealt_to_turrets(self) -> int:
        return self._dto["damageDealtToTurrets"]

    @property
    def altars_neutralized(self) -> int:
        return self._dto["altarsNeutralized"]

    @property
    def physical_damage_dealt_to_champions(self) -> int:
        return self._dto["physicalDamageDealtToChampions"]

    @property
    def gold_spent(self) -> int:
        return self._dto["goldSpent"]

    @property
    def true_damage_dealt(self) -> int:
        return self._dto["trueDamageDealt"]

    @property
    def true_damage_dealt_to_champions(self) -> int:
        return self._dto["trueDamageDealtToChampions"]

    @property
    def id(self) -> int:
        return self._dto["participantId"]

    @property
    def penta_kills(self) -> int:
        return self._dto["pentaKills"]

    @property
    def total_heal(self) -> int:
        return self._dto["totalHeal"]

    @property
    def total_minions_killed(self) -> int:
        return self._dto["totalMinionsKilled"]

    @property
    def first_blood_kill(self) -> bool:
        return self._dto["firstBloodKill"]

    @property
    def node_capture(self) -> int:
        return self._dto["nodeCapture"]

    @property
    def largest_multi_kill(self) -> int:
        return self._dto["largestMultKill"]

    @property
    def sight_wards_bought_in_game(self) -> int:
        return self._dto["sightWardsBoughtInGame"]

    @property
    def total_damage_dealt_to_champions(self) -> int:
        return self._dto["totalDamageDealtToChampions"]

    @property
    def total_units_healed(self) -> int:
        return self._dto["totalUnitsHealed"]

    @property
    def inhibitor_kills(self) -> int:
        return self._dto["inhibitorKills"]

    @property
    def total_score_rank(self) -> int:
        return self._dto["totalScoreRank"]

    @property
    def total_damage_taken(self) -> int:
        return self._dto["totalDamageTaken"]

    @property
    def killing_sprees(self) -> int:
        return self._dto["killingSprees"]

    @property
    def time_CCing_others(self) -> int:
        return self._dto["timeCCingOthers"]

    @property
    def physical_damage_taken(self) -> int:
        return self._dto["physicalDamageTaken"]


class ParticipantData(CoreData):
    _renamed = {"id": "participantId", "side": "teamId", "summoner_spell_d_id": "spell1Id", "summoner_spell_f_id": "spell2Id", "champion_id": "championId", "rank_last_season":"highestAchievedSeasonTier"}

    @property
    def stats(self) -> ParticipantStatsData:
        return ParticipantStatsData.from_dto(self._dto["stats"])

    @property
    def id(self) -> int:
        return self._dto["participantId"]

    @property
    def runes(self) -> List[int]:
        return self._dto["runes"]

    @property
    def masteries(self) -> List[int]:
        return self._dto["masteries"]

    @property
    def timeline(self) -> ParticipantTimelineData:
        return ParticipantTimelineData.from_dto(self._dto["timeline"])

    @property
    def side(self) -> int:
        return self._dto["teamId"]

    @property
    def summoner_spell_d_id(self) -> int:
        return self._dto["spell1Id"]

    @property
    def summoner_spell_f_id(self) -> int:
        return self._dto["spell2Id"]

    @property
    def rank_last_season(self) -> Tuple[str, int]:
        return self._dto["highestAchievedSeasonTier"]

    @property
    def champion_id(self) -> int:
        return self._dto["championId"]


class PlayerData(CoreData):
    _renamed = {"current_platform_id": "currentPlatformId", "summoner_name": "summonerName", "summoner_id": "summonerId", "match_history_uri": "matchHistoryUri", "platform_id": "platformId", "current_account_id": "currentAccountId", "profile_icon_id": "profileIcon", "account_id": "accountId"}

    @property
    def current_platform_id(self) -> int:
        return self._dto["currentPlatformId"]

    @property
    def summoner_id(self) -> int:
        return self._dto["summonerId"]

    @property
    def summoner_name(self) -> str:
        return self._dto["summonerName"]

    @property
    def match_history_uri(self) -> str:
        return self._dto["matchHistoryUri"]

    @property
    def platform_id(self) -> int:
        return self._dto["platformId"]

    @property
    def profile_icon_id(self) -> int:
        return self._dto["profileIcon"]

    @property
    def current_account_id(self) -> int:
        return self._dto["currentAccountId"]

    @property
    def account_id(self) -> int:
        return self._dto["accountId"]


class ParticipantIdentityData(CoreData):
    _renamed = {"id": "participantId"}

    @property
    def player(self) -> PlayerData:
        return PlayerData.from_dto(self._dto["player"])

    @property
    def id(self) -> int:
        return self._dto["participantId"]


class TeamData(CoreData):
    _renamed = {"first_dragon": "firstDragon", "first_inhibitor": "firstInhibitor", "first_rift_herald": "firstRiftHerald", "first_baron": "firstBaron", "first_tower": "firstTower", "first_blood": "firstBlood", "baron_kills": "baronKills", "rift_herald_kills": "riftHeraldKills", "vilemaw_kills": "vilemawKills", "inhibitor_kills": "inhibitorKills", "tower_kills": "towerKills", "dragon_kills": "dragonKills", "dominion_victory_score": "dominionVictoryScore", "side": "teamId"}

    @property
    def first_dragon(self) -> bool:
        return self._dto["firstDragon"]

    @property
    def first_inhibitor(self) -> bool:
        return self._dto["firstInhibitor"]

    @property
    def first_rift_herald(self) -> bool:
        return self._dto["firstRiftHerald"]

    @property
    def first_baron(self) -> bool:
        return self._dto["firstBaron"]

    @property
    def first_tower(self) -> bool:
        return self._dto["firstTower"]

    @property
    def first_blood(self) -> bool:
        return self._dto["firstBlood"]

    @property
    def bans(self) -> List[int]:
        return [ban["championId"] for ban in self._dto["bans"]]

    @property
    def baron_kills(self) -> int:
        return self._dto["baronKills"]

    @property
    def rift_herald_kills(self) -> int:
        return self._dto["riftHeraldKills"]

    @property
    def vilemaw_kills(self) -> int:
        return self._dto["vilemawKills"]

    @property
    def inhibitor_kills(self) -> int:
        return self._dto["inhibitorKills"]

    @property
    def tower_kills(self) -> int:
        return self._dto["towerKills"]

    @property
    def dragon_kills(self) -> int:
        return self._dto["dragonKills"]

    @property
    def side(self) -> int:
        return self._dto["teamId"]

    @property
    def dominion_victory_score(self) -> int:
        return self._dto["dominionVictoryScore"]

    @property
    def win(self) -> bool:
        return self._dto["win"]


class MatchReferenceData(CoreData):
    _renamed = {"account_id": "accountId", "season_id": "season", "queue_id": "queue", "id": "gameId", "platform_id": "platformId", "champion_id": "champion", "creation": "timestamp"}

    @property
    def region(self) -> int:
        return self._dto["region"]

    @property
    def account_id(self) -> int:
        return self._dto["accountId"]

    @property
    def id(self) -> int:
        return self._dto["gameId"]

    @property
    def season_id(self) -> int:
        return self._dto["season"]

    @property
    def queue_id(self) -> int:
        return self._dto["queue"]

    @property
    def platform_id(self) -> int:
        return self._dto["platformId"]

    @property
    def lane(self) -> str:
        return self._dto["lane"]

    @property
    def role(self) -> str:
        return self._dto["role"]

    @property
    def champion_id(self) -> int:
        return self._dto["champion"]

    @property
    def creation(self) -> int:
        return self._dto["timestamp"]


class MatchData(CoreData):
    _dto_type = dto.MatchDto
    _renamed = {"season_id": "seasonId", "queue_id": "queueId", "id": "gameId", "participant_identities": "participantIdentities", "version": "gameVersion", "platform_id": "platformId", "mode": "gameMode", "map_id": "mapId", "type": "gameType", "duration": "gameDuration", "creation": "gameCreation"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def id(self) -> int:
        return self._dto["gameId"]

    @property
    def season_id(self) -> int:
        return self._dto["seasonId"]

    @property
    def queue_id(self) -> int:
        return self._dto["queueId"]

    @property
    def participants(self) -> List[ParticipantData]:
        return [ParticipantData.from_dto(p) for p in self._dto["participants"]]

    @property
    def participant_identities(self) -> List[ParticipantIdentityData]:
        return [ParticipantIdentityData.from_dto(p) for p in self._dto["participantIdentities"]]

    @property
    def teams(self) -> List[TeamData]:
        return [TeamData.from_dto(t) for t in self._dto["teams"]]

    @property
    def version(self) -> str:
        return self._dto["gameVersion"]

    @property
    def platform_id(self) -> int:
        return self._dto["platformId"]

    @property
    def mode(self) -> str:
        return self._dto["gameMode"]

    @property
    def map_id(self) -> int:
        return self._dto["mapId"]

    @property
    def type(self) -> str:
        return self._dto["gameType"]

    @property
    def duration(self) -> int:
        return self._dto["gameDuration"]

    @property
    def creation(self) -> int:
        return self._dto["gameCreation"]


##############
# Core Types #
##############


class MatchHistory(CassiopeiaGhostList):
    _data_types = {MatchListData}

    def __init__(self, *args, summoner: Union[Summoner, str, int] = None, account_id: int = None, region: Union[Region, str]):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region}
        if account_id is not None and summoner is None:
            summoner = Summoner(account=account_id, region=region)
        elif isinstance(summoner, int):
            summoner = Summoner(id=summoner, region=region)
        elif isinstance(summoner, str):
            summoner = Summoner(name=summoner, region=region)
        assert isinstance(summoner, Summoner)
        self.__class__.summoner.fget._lazy_set(self, summoner)
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        query = {"platform": self.platform, "account.id": self.summoner.account.id}
        return query

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ..transformers.match import MatchTransformer
        SearchableList.__init__(self, [MatchTransformer.match_reference_data_to_core(None, i) for i in data])
        # We have a summoner object (probably) already created, and if one was passed in then this is pretty crucial to do:
        # Put the summoner object that this match history was instantiated with into the participant[0] so that searchable
        # list syntax on e.g. the name will work without loading the summoner.
        # For example:
        #    summoner = Summoner(name=name, account=account, id=id, region=region)
        #    match = summoner.match_history[0]
        #    p = match.participants[name]  # This will work without loading the summoner because the name was prvided
        for match in self:
            match.participants[0].__class__.summoner.fget._lazy_set(match.participants[0], self.summoner)
        super().__load_hook__(load_group, data)

    @lazy_property
    def summoner(self) -> Summoner:
        return Summoner(account=self._data[MatchListData].account_id)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MatchListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @lazy_property
    def queues(self) -> Set[Queue]:
        return {Queue(q) for q in self._data[MatchListData].queues}

    @lazy_property
    def seasons(self) -> Set[Season]:
        return {Season(s) for s in self._data[MatchListData].seasons}

    @lazy_property
    def champions(self) -> Set[Champion]:
        return {Champion(id=cid) for cid in self._data[MatchListData].champion_ids}

    @property
    def begin_index(self) -> int:
        return self._data[MatchListData].begin_index

    @property
    def end_index(self) -> int:
        return self._data[MatchListData].end_index

    @property
    def begin_time(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self._data[MatchListData].begin_time / 1000)

    @property
    def end_time(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self._data[MatchListData].end_time / 1000)


class Postion(CassiopeiaObject):
    _data_types = {PositionData}

    @property
    def x(self) -> int:
        return self._data[PositionData].x

    @property
    def y(self) -> int:
        return self._data[PositionData].y


@searchable({str: ["event_type", "tower_type", "ascended_type", "ward_type", "monster_type", "type", "monster_sub_type", "lane_type", "building_type"]})
class Event(CassiopeiaObject):
    _data_types = {EventData}

    @property
    def event_type(self) -> str:
        return self._data[EventData].event_type

    @property
    def tower_type(self) -> str:
        return self._data[EventData].tower_type

    @property
    def team_id(self) -> int:
        return self._data[EventData].team_id

    @property
    def ascended_type(self) -> str:
        return self._data[EventData].ascended_type

    @property
    def killer_id(self) -> int:
        return self._data[EventData].killer_id

    @property
    def level_up_type(self) -> str:
        return self._data[EventData].level_up_type

    @property
    def point_captured(self) -> str:
        return self._data[EventData].point_captured

    @property
    def assisting_participant_ids(self) -> List[int]:
        return self._data[EventData].assisting_participant_ids

    @property
    def ward_type(self) -> str:
        return self._data[EventData].ward_type

    @property
    def monster_type(self) -> str:
        return self._data[EventData].monster_type

    @property
    def type(self) -> List[str]:
        """Legal values: CHAMPION_KILL, WARD_PLACED, WARD_KILL, BUILDING_KILL, ELITE_MONSTER_KILL, ITEM_PURCHASED, ITEM_SOLD, ITEM_DESTROYED, ITEM_UNDO, SKILL_LEVEL_UP, ASCENDED_EVENT, CAPTURE_POINT, PORO_KING_SUMMON"""
        return self._data[EventData].type

    @property
    def skill_slot(self) -> int:
        return self._data[EventData].skill_slot

    @property
    def victim_id(self) -> int:
        return self._data[EventData].victim_id

    @property
    def timestamp(self) -> int:
        return self._data[EventData].timestamp

    @property
    def after_id(self) -> int:
        return self._data[EventData].after_id

    @property
    def monster_sub_type(self) -> str:
        return self._data[EventData].monster_sub_type

    @property
    def lane_type(self) -> str:
        return self._data[EventData].lane_type

    @property
    def item_id(self) -> int:
        return self._data[EventData].items_id

    @property
    def participant_id(self) -> int:
        return self._data[EventData].participant_id

    @property
    def building_type(self) -> str:
        return self._data[EventData].building_type

    @property
    def creator_id(self) -> int:
        return self._data[EventData].creator_id

    @property
    def position(self) -> PositionData:
        return self._data[EventData].position

    @property
    def before_id(self) -> int:
        return self._data[EventData].before_id


class ParticipantFrame(CassiopeiaObject):
    _data_types = {ParticipantFrameData}

    @property
    def total_gold(self) -> int:
        return self._data[ParticipantFrameData].total_gold

    @property
    def team_score(self) -> int:
        return self._data[ParticipantFrameData].team_score

    @property
    def participant_id(self) -> int:
        return self._data[ParticipantFrameData].participant_id

    @property
    def level(self) -> int:
        return self._data[ParticipantFrameData].level

    @property
    def current_gold(self) -> int:
        return self._data[ParticipantFrameData].current_gold

    @property
    def minions_killed(self) -> int:
        return self._data[ParticipantFrameData].minions_killed

    @property
    def dominion_score(self) -> int:
        return self._data[ParticipantFrameData].dominion_score

    @property
    def position(self) -> PositionData:
        return self._data[ParticipantFrameData].position

    @property
    def xp(self) -> int:
        return self._data[ParticipantFrameData].xp

    @property
    def jungle_minions_killed(self) -> int:
        return self._data[ParticipantFrameData].jungle_minions_killed


class Frame(CassiopeiaObject):
    _data_types = {FrameData}

    @property
    def timestamp(self) -> int:
        return self._data[FrameData].timestamp

    @property
    def participant_frames(self) -> Dict[int, ParticipantFrame]:
        return SearchableDictionary({k: ParticipantFrame.from_data(frame) for k, frame in self._data[FrameData].participant_frames.items()})

    @property
    def events(self) -> List[Event]:
        return SearchableList([Event(event) for event in self._data[FrameData].events])


class Timeline(CassiopeiaGhost):
    _data_types = {TimelineData}

    def __init__(self, *, id: int = None, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region, "id": id}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "matchId": self.id}

    @property
    def id(self):
        return self._data[TimelineData].id

    @property
    def region(self) -> Region:
        return Region(self._data[TimelineData].region)

    @property
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(TimelineData)
    @ghost_load_on(KeyError)
    def frames(self) -> List[Frame]:
        return SearchableList([Frame.from_data(frame) for frame in self._data[TimelineData].frame])

    @CassiopeiaGhost.property(TimelineData)
    @ghost_load_on(KeyError)
    def frame_interval(self) -> int:
        return self._data[TimelineData].frame_interval


class ParticipantTimeline(CassiopeiaObject):
    _data_types = {ParticipantTimelineData}

    # TODO Add lane and role enums and make things searchable by them
    @property
    def lane(self) -> str:
        return self._data[ParticipantTimelineData].lane

    @property
    def role(self) -> str:
        return self._data[ParticipantTimelineData].role

    @property
    def id(self) -> int:
        return self._data[ParticipantTimelineData].id

    @property
    def cs_diff_per_min_deltas(self) -> Dict[str, float]:
        return self._data[ParticipantTimelineData].cs_diff_per_min_deltas

    @property
    def gold_per_min_deltas(self) -> Dict[str, float]:
        return self._data[ParticipantTimelineData].gold_per_min_deltas

    @property
    def xp_diff_per_min_deltas(self) -> Dict[str, float]:
        return self._data[ParticipantTimelineData].xp_diff_per_min_deltas

    @property
    def creeps_per_min_deltas(self) -> Dict[str, float]:
        return self._data[ParticipantTimelineData].creeps_per_min_deltas

    @property
    def xp_per_min_deltas(self) -> Dict[str, float]:
        return self._data[ParticipantTimelineData].xp_per_min_deltas

    @property
    def damage_taken_per_min_deltas(self) -> Dict[str, float]:
        return self._data[ParticipantTimelineData].damage_taken_per_min_deltas

    @property
    def damage_taken_diff_per_min_deltas(self) -> Dict[str, float]:
        return self._data[ParticipantTimelineData].damage_taken_diff_per_min_deltas


@searchable({str: ["items"], Item: ["items"]})
class ParticipantStats(CassiopeiaObject):
    _data_types = {ParticipantStatsData}

    @property
    def physical_damage_dealt(self) -> int:
        return self._data[ParticipantStatsData].physical_damage_dealt

    @property
    def magic_damage_dealt(self) -> int:
        return self._data[ParticipantStatsData].magic_damage_dealt

    @property
    def neutral_minions_killed_team_jungle(self) -> int:
        return self._data[ParticipantStatsData].neutral_minions_killed_team_jungle

    @property
    def total_player_score(self) -> int:
        return self._data[ParticipantStatsData].total_player_score

    @property
    def deaths(self) -> int:
        return self._data[ParticipantStatsData].deaths

    @property
    def win(self) -> bool:
        return self._data[ParticipantStatsData].win

    @property
    def neutral_minions_killed_enemy_jungle(self) -> int:
        return self._data[ParticipantStatsData].neutral_minions_killed_enemy_jungle

    @property
    def altars_captured(self) -> int:
        return self._data[ParticipantStatsData].altars_captured

    @property
    def largest_critical_strike(self) -> int:
        return self._data[ParticipantStatsData].largest_critical_strike

    @property
    def total_damage_dealt(self) -> int:
        return self._data[ParticipantStatsData].total_damage_dealt

    @property
    def magic_damage_dealt_to_champions(self) -> int:
        return self._data[ParticipantStatsData].magic_damage_dealt_to_champions

    @property
    def vision_wards_bought_in_game(self) -> int:
        return self._data[ParticipantStatsData].vision_wards_bought_in_game

    @property
    def damage_dealt_to_objectives(self) -> int:
        return self._data[ParticipantStatsData].damage_dealt_to_objectives

    @property
    def largest_killing_spree(self) -> int:
        return self._data[ParticipantStatsData].largest_killing_spree

    @property
    def quadra_kills(self) -> int:
        return self._data[ParticipantStatsData].quadra_kills

    @property
    def team_objective(self) -> int:
        return self._data[ParticipantStatsData].team_objective

    @property
    def total_time_crowd_control_dealt(self) -> int:
        return self._data[ParticipantStatsData].total_time_crowd_control_dealt

    @property
    def longest_time_spent_living(self) -> int:
        return self._data[ParticipantStatsData].longest_time_spent_living

    @property
    def wards_killed(self) -> int:
        return self._data[ParticipantStatsData].wards_killed

    @property
    def first_tower_assist(self) -> bool:
        return self._data[ParticipantStatsData].first_tower_assist

    @property
    def first_tower_kill(self) -> bool:
        return self._data[ParticipantStatsData].first_tower_kill

    @property
    def items(self) -> List[Item]:
        ids = [self._data[ParticipantStatsData].item0,
               self._data[ParticipantStatsData].item1,
               self._data[ParticipantStatsData].item2,
               self._data[ParticipantStatsData].item3,
               self._data[ParticipantStatsData].item4,
               self._data[ParticipantStatsData].item5,
               self._data[ParticipantStatsData].item6
        ]
        return SearchableList([Item(id=id) for id in ids if id])

    @property
    def first_blood_assist(self) -> bool:
        return self._data[ParticipantStatsData].first_blood_assist

    @property
    def vision_score(self) -> int:
        return self._data[ParticipantStatsData].vision_score

    @property
    def wards_placed(self) -> int:
        return self._data[ParticipantStatsData].wards_placed

    @property
    def turret_kills(self) -> int:
        return self._data[ParticipantStatsData].turret_kills

    @property
    def triple_kills(self) -> int:
        return self._data[ParticipantStatsData].triple_kills

    @property
    def damage_self_mitigated(self) -> int:
        return self._data[ParticipantStatsData].damage_self_mitigated

    @property
    def champion_level(self) -> int:
        return self._data[ParticipantStatsData].champion_level

    @property
    def node_neutralize_assist(self) -> int:
        return self._data[ParticipantStatsData].node_neutralize_assist

    @property
    def first_inhibitor_kill(self) -> bool:
        return self._data[ParticipantStatsData].first_inhibitor_kill

    @property
    def gold_earned(self) -> int:
        return self._data[ParticipantStatsData].gold_earned

    @property
    def magical_damage_taken(self) -> int:
        return self._data[ParticipantStatsData].magical_damage_taken

    @property
    def kills(self) -> int:
        return self._data[ParticipantStatsData].kills

    @property
    def double_kills(self) -> int:
        return self._data[ParticipantStatsData].double_kills

    @property
    def node_capture_assist(self) -> int:
        return self._data[ParticipantStatsData].node_capture_assist

    @property
    def true_damage_taken(self) -> int:
        return self._data[ParticipantStatsData].true_damage_taken

    @property
    def node_neutralize(self) -> int:
        return self._data[ParticipantStatsData].node_neutralize

    @property
    def first_inhibitor_assist(self) -> bool:
        return self._data[ParticipantStatsData].first_inhibitor_assist

    @property
    def assists(self) -> int:
        return self._data[ParticipantStatsData].assists

    @property
    def unreal_kills(self) -> int:
        return self._data[ParticipantStatsData].unreal_kills

    @property
    def neutral_minions_killed(self) -> int:
        return self._data[ParticipantStatsData].neutral_minions_killed

    @property
    def objective_player_score(self) -> int:
        return self._data[ParticipantStatsData].objective_player_score

    @property
    def combat_player_score(self) -> int:
        return self._data[ParticipantStatsData].combat_player_score

    @property
    def damage_dealt_to_turrets(self) -> int:
        return self._data[ParticipantStatsData].damage_dealt_to_turrets

    @property
    def altars_neutralized(self) -> int:
        return self._data[ParticipantStatsData].altars_neutralized

    @property
    def physical_damage_dealt_to_champions(self) -> int:
        return self._data[ParticipantStatsData].physical_damage_dealt_to_champions

    @property
    def gold_spent(self) -> int:
        return self._data[ParticipantStatsData].gold_spent

    @property
    def true_damage_dealt(self) -> int:
        return self._data[ParticipantStatsData].true_damage_dealt

    @property
    def true_damage_dealt_to_champions(self) -> int:
        return self._data[ParticipantStatsData].true_damage_dealt_to_champions

    @property
    def id(self) -> int:
        return self._data[ParticipantStatsData].id

    @property
    def penta_kills(self) -> int:
        return self._data[ParticipantStatsData].penta_kills

    @property
    def total_heal(self) -> int:
        return self._data[ParticipantStatsData].total_heal

    @property
    def total_minions_killed(self) -> int:
        return self._data[ParticipantStatsData].total_minions_killed

    @property
    def first_blood_kill(self) -> bool:
        return self._data[ParticipantStatsData].first_blood_kill

    @property
    def node_capture(self) -> int:
        return self._data[ParticipantStatsData].node_capture

    @property
    def largest_multi_kill(self) -> int:
        return self._data[ParticipantStatsData].largest_multi_kill

    @property
    def sight_wards_bought_in_game(self) -> int:
        return self._data[ParticipantStatsData].sight_wards_bought_in_game

    @property
    def total_damage_dealt_to_champions(self) -> int:
        return self._data[ParticipantStatsData].total_damage_dealt_to_champions

    @property
    def total_units_healed(self) -> int:
        return self._data[ParticipantStatsData].total_units_healed

    @property
    def inhibitor_kills(self) -> int:
        return self._data[ParticipantStatsData].inhibitor_kills

    @property
    def total_score_rank(self) -> int:
        return self._data[ParticipantStatsData].total_score_rank

    @property
    def total_damage_taken(self) -> int:
        return self._data[ParticipantStatsData].total_damage_taken

    @property
    def killing_sprees(self) -> int:
        return self._data[ParticipantStatsData].killing_sprees

    @property
    def time_CCing_others(self) -> int:
        return self._data[ParticipantStatsData].time_CCing_others

    @property
    def physical_damage_taken(self) -> int:
        return self._data[ParticipantStatsData].physical_damage_taken


def load_match_on_keyerror(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except KeyError:  # teamId
            # The match has only partially loaded this participant and it doesn't have all it's data, so load the full match
            self._Participant__match.__load__(MatchData)
            self._Participant__match._Ghost__set_loaded(MatchData)
            for participant in self._Participant__match.participants:
                if participant.summoner.name == self.summoner.name:
                    self._data[ParticipantData] = participant._data[ParticipantData]
                    return method(self, *args, **kwargs)
        return method(self, *args, **kwargs)
    return wrapper


@searchable({str: ["summoner", "champion", "stats", "runes", "masteries", "side", "summoner_spell_d", "summoner_spell_f"], Summoner: ["summoner"], Champion: ["champion"], Side: ["side"], Rune: ["runes"], Mastery: ["masteries"], SummonerSpell: ["summoner_spell_d", "summoner_spell_f"]})
class Participant(CassiopeiaObject):
    _data_types = {ParticipantData, PlayerData}

    @classmethod
    def from_data(cls, data: CoreData, match: "Match"):
        self = super().from_data(data)
        self.__match = match
        return self

    @lazy_property
    @load_match_on_keyerror
    def stats(self) -> ParticipantStats:
        return ParticipantStats.from_data(self._data[ParticipantData].stats)

    @property
    def id(self) -> int:
        return self._data[ParticipantData].id

    @lazy_property
    @load_match_on_keyerror
    def runes(self) -> List["Rune"]:
        return self._data[ParticipantData].runes

    @lazy_property
    @load_match_on_keyerror
    def masteries(self) -> List["Mastery"]:
        return self._data[ParticipantData].masteries

    @lazy_property
    @load_match_on_keyerror
    def timeline(self) -> ParticipantTimeline:
        return ParticipantTimeline.from_data(self._data[ParticipantData].timeline)

    @lazy_property
    @load_match_on_keyerror
    def side(self) -> Side:
        return Side(self._data[ParticipantData].side)

    @lazy_property
    @load_match_on_keyerror
    def summoner_spell_d(self) -> SummonerSpell:
        return SummonerSpell(id=self._data[ParticipantData].summoner_spell_d_id)

    @lazy_property
    @load_match_on_keyerror
    def summoner_spell_f(self) -> SummonerSpell:
        return SummonerSpell(id=self._data[ParticipantData].summoner_spell_f_id)

    @lazy_property
    @load_match_on_keyerror
    def rank_last_season(self) -> Tuple[Tier, Division]:
        return Tier(self._data[ParticipantData].rank_last_season[0]), Division(self._data[ParticipantData].rank_last_season[1])

    @lazy_property
    @load_match_on_keyerror
    def champion(self) -> "Champion":
        from .staticdata.champion import Champion
        return Champion(id=self._data[ParticipantData].champion_id)

    # All the Player data from ParticipantIdentities.player is contained in the Summoner class.
    # The non-current accountId and platformId should never be relevant/used, and can be deleted from our type system.
    #   See: https://discussion.developer.riotgames.com/questions/1713/is-there-any-scenario-where-accountid-should-be-us.html
    @lazy_property
    def summoner(self) -> "Summoner":
        kwargs = {}
        try:
            kwargs["id"] = self._data[PlayerData].summoner_id
        except KeyError:
            pass
        try:
            kwargs["name"] = self._data[PlayerData].summoner_name
        except KeyError:
            pass
        from .summoner import Summoner, ProfileIcon
        kwargs["account"] = self._data[PlayerData].account_id
        kwargs["region"] = Platform(self._data[PlayerData].current_platform_id).region
        # TODO Add profile icon
        #try:
        #    kwargs["profile_icon"] = ProfileIcon(id=self._data[PlayerData].profile_icon_id)
        #except KeyError:
        #    pass
        return Summoner(**kwargs)

    @property
    def team(self) -> "Team":
        if self.side == Side.blue:
            return self.__match.blue_team
        else:
            return self.__match.red_team


@searchable({str: ["participants"], bool: ["win"]})
class Team(CassiopeiaObject):
    _data_types = {TeamData}

    @classmethod
    def from_data(cls, data: CoreData, participants: List[Participant]):
        self = super().from_data(data)
        self.__participants = participants
        return self

    @property
    def first_dragon(self) -> bool:
        return self._data[TeamData].first_dragon

    @property
    def first_inhibitor(self) -> bool:
        return self._data[TeamData].first_inhibitor

    @property
    def first_rift_herald(self) -> bool:
        return self._data[TeamData].first_rift_herald

    @property
    def first_baron(self) -> bool:
        return self._data[TeamData].first_baron

    @property
    def first_tower(self) -> bool:
        return self._data[TeamData].first_tower

    @property
    def first_blood(self) -> bool:
        return self._data[TeamData].first_blood

    @property
    def bans(self) -> List["Champion"]:
        return [Champion(id=champion_id) for champion_id in self._data[TeamData].bans]

    @property
    def baron_kills(self) -> int:
        return self._data[TeamData].baron_kills

    @property
    def rift_herald_kills(self) -> int:
        return self._data[TeamData].rift_herald_kills

    @property
    def vilemaw_kills(self) -> int:
        return self._data[TeamData].vilemaw_kills

    @property
    def inhibitor_kills(self) -> int:
        return self._data[TeamData].inhibitor_kills

    @property
    def tower_kills(self) -> int:
        return self._data[TeamData].tower_kills

    @property
    def dragon_kills(self) -> int:
        return self._data[TeamData].dragon_kills

    @property
    def side(self) -> Side:
        return Side(self._data[TeamData].side)

    @property
    def dominion_victory_score(self) -> int:
        return self._data[TeamData].dominion_victory_score

    @property
    def win(self) -> bool:
        return self._data[TeamData].win != "Fail"

    @property
    def participants(self) -> List[Participant]:
        return self.__participants


@searchable({str: ["participants", "region", "platform", "season", "queue", "mode", "map", "type"], Region: ["region"], Platform: ["platform"], Season: ["season"], Queue: ["queue"], GameMode: ["mode"], Map: ["map"], GameType: ["type"], Item: ["participants"], Champion: ["participants"]})
class Match(CassiopeiaGhost):
    _data_types = {MatchData}

    def __init__(self, *, id: int = None, region: Union[Region, str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        kwargs = {"region": region, "id": id}
        super().__init__(**kwargs)
        self.__participants = []  # For lazy-loading the participants in a special way

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "gameId": self.id}

    @classmethod
    def from_match_reference(cls, ref):
        # TODO Somehow put in ref.lane, ref.role, and ref.champion_id
        participant = {"participantId": 0, "championId": ref.champion_id}
        player = {"participantId": 0, "accountId": ref.account_id, "currentPlatformId": ref.platform_id}
        instance = cls(id=ref.id, region=ref.region)
        instance(season_id=ref.season_id, queue_id=ref.queue_id, creation=ref.creation)
        instance._data[MatchData]._dto["participants"] = [participant]
        instance._data[MatchData]._dto["participantIdentities"] = [{"participantId": 0, "player": player}]
        return instance

    @lazy_property
    def region(self) -> Region:
        """The region for this match."""
        return Region(self._data[MatchData].region)

    @property
    def platform(self) -> Platform:
        """The platform for this match."""
        return self.region.platform

    @property
    def id(self) -> int:
        return self._data[MatchData].id

    @lazy_property
    def timeline(self) -> Timeline:
        return Timeline(id=self.id, region=self.region.value)

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def season(self) -> Season:
        return Season.from_id(self._data[MatchData].season_id)

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def queue(self) -> Queue:
        return Queue.from_id(self._data[MatchData].queue_id)

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    # This method is lazy-loaded in a special way because of its unique behavior
    def participants(self) -> List[Participant]:
        # This is a complicated function because we don't want to load the particpants if the only one the user cares about is the one loaded from a match ref

        def construct_participant(participant_data, participant_identities):
            """A helper function for creating a participant from participant, participant identity, and player data."""
            for pidentity in participant_identities:
                if pidentity.id == participant_data.id:
                    participant = Participant.from_data(ParticipantData.from_dto({}), match=self)
                    participant._data[ParticipantData] = participant_data
                    participant._data[PlayerData] = pidentity.player
                    return participant

        def generate_participants(match):
            # If a participant was provided from a matchref, yield that first
            yielded_one = False
            if len(match._data[MatchData].participants) == 1:
                yielded_one = True
                try:
                    yield match.__participants[0]
                except IndexError:
                    p = match._data[MatchData].participants[0]
                    participant = construct_participant(p, match._data[MatchData].participant_identities)
                    match.__participants.append(participant)
                    yield participant

            # Create all the participants if any haven't been created yet.
            # Note that it's important to overwrite the one from the matchref if it was loaded because we have more data after we load the full match.
            if yielded_one or len(match.__participants) < len(match._data[MatchData].participants):
                if not match._Ghost__is_loaded(MatchData):
                    match.__load__(MatchData)
                    match._Ghost__set_loaded(MatchData)  # __load__ doesn't trigger __set_loaded. is this a "bug"?
                match.__participants = [None for _ in match._data[MatchData].participants]
                for i, p in enumerate(match._data[MatchData].participants):
                    participant = construct_participant(p, match._data[MatchData].participant_identities)
                    match.__participants[i] = participant

            # Yield the rest of the participants
            for participant in match.__participants[yielded_one:]:
                yield participant

        return SearchableLazyList(generate_participants(self))

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def teams(self) -> List[Team]:
        return [Team.from_data(t, participants=[p for p in self.participants if p.side.value == self._data[MatchData].teams[i].side]) for i, t in enumerate(self._data[MatchData].teams)]

    @property
    def red_team(self) -> Team:
        if self.teams[0].side is Side.red:
            return self.teams[0]
        else:
            return self.teams[1]

    @property
    def blue_team(self) -> Team:
        if self.teams[0].side is Side.blue:
            return self.teams[0]
        else:
            return self.teams[1]

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    def version(self) -> str:
        return self._data[MatchData].version

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def mode(self) -> GameMode:
        return GameMode(self._data[MatchData].mode)

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def map(self) -> Map:
        return Map(self._data[MatchData].map_id)

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def type(self) -> GameType:
        return GameType(self._data[MatchData].type)

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def duration(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self._data[MatchData].duration)

    @CassiopeiaGhost.property(MatchData)
    @ghost_load_on(KeyError)
    @lazy
    def creation(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self._data[MatchData].creation / 1000)



from .staticdata import Champion, Champions, Rune, Mastery, Item, Map, SummonerSpell, Realms, ProfileIcon, ProfileIcons, LanguageStrings, Locales, Versions, Masteries, Runes, SummonerSpells, Maps, Items
from .summoner import Summoner, Account
from .championmastery import ChampionMastery, ChampionMasteries
from .runepage import RunePage, RunePages
from .masterypage import MasteryPage, MasteryPages
from .match import Match, MatchHistory
from .spectator import CurrentMatch, FeaturedMatches
from .status import ShardStatus
from .league import ChallengerLeague, MasterLeague, Leagues


from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query
from .common import RiotAPIService, APINotFoundError
from ...data import Platform
from ...dto.championmastery import ChampionMasteryDto, ChampionMasteryListDto, ChampionMasteryScoreDto
from ...dto.summoner import SummonerDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


class ChampionMasteryAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_champion_mastery_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(int).or_("summoner.account.id").as_(int).or_("summoner.name").as_(str).also. \
        has("champion.id").as_(int).or_("champion.name").as_(str)

    @get.register(ChampionMasteryDto)
    @validate_query(_validate_get_champion_mastery_query, convert_region_to_platform)
    def get_champion_mastery(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionMasteryDto:
        # TODO Is this still what we want to do?
        if "summoner.id" not in query:
            for k in ["summoner.account.id", "summoner.account_id", "summoner.accountId"]:
                if k in query:
                    summoner_query = {"platform": query["platform"], "account.id": query[k]}
                    break
            for k in ["summoner.name", "summoner_name", "summonerName"]:
                if k in query:
                    summoner_query = {"platform": query["platform"], "name": query[k]}
                    break
            summoner_dto = context[PipelineContext.Keys.PIPELINE].get(SummonerDto, query=summoner_query)
            query["summoner.id"] = summoner_dto["id"]

        url = "https://{platform}.api.riotgames.com/lol/champion-mastery/v3/champion-masteries/by-summoner/{summonerId}/by-champion/{championId}".format(platform=query["platform"].value.lower(), summonerId=query["summoner.id"], championId=query["champion.id"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "champion-masteries/by-summoner/summonerId/by-champion/championId"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["summonerId"] = query["summoner.id"]
        return ChampionMasteryDto(data)

    _validate_get_many_champion_mastery_query = Query. \
        has("platform").as_(Platform).also. \
        has("summonerId").as_(int).also. \
        has("championIds").as_(Iterable)

    @get_many.register(ChampionMasteryDto)
    def get_many_champion_mastery(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionMasteryDto, None, None]:
        ChampionMasteryAPI._validate_get_many_champion_mastery_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/champion-mastery/v3/champion-masteries/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=query["summonerId"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "champion-masteries/by-summoner/summonerId"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        masteries = {
            mastery["championId"]: mastery for mastery in data
        }

        def generator():
            for id in query["championIds"]:
                try:
                    mastery = masteries[id]
                except KeyError as error:
                    raise NotFoundError("Summoner has no mastery on champion with id \"{id}\"".format(id=id)) from error

                mastery["summonerId"] = query["summonerId"]
                mastery["region"] = query["platform"].region.value
                yield ChampionMasteryDto(mastery)

        return generator()

    _validate_get_champion_mastery_list_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(int)

    @get.register(ChampionMasteryListDto)
    def get_champion_mastery_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionMasteryListDto:
        ChampionMasteryAPI._validate_get_champion_mastery_list_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/champion-mastery/v3/champion-masteries/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=query["summoner.id"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "champion-masteries/by-summoner/summonerId"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        return ChampionMasteryListDto({
            "masteries": data,
            "summonerId": query["summoner.id"],
            "region": query["platform"].region.value
        })

    _validate_get_many_champion_mastery_list_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.ids").as_(Iterable)

    @get_many.register(ChampionMasteryListDto)
    def get_many_champion_mastery_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionMasteryListDto, None, None]:
        ChampionMasteryAPI._validate_get_many_champion_mastery_list_query(query, context)

        def generator():
            for summoner_id in query["summoner.ids"]:
                url = "https://{platform}.api.riotgames.com/lol/champion-mastery/v3/champion-masteries/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=summoner_id)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "champion-masteries/by-summoner/summonerId"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                yield ChampionMasteryListDto({
                    "masteries": data,
                    "summonerId": summoner_id,
                    "region": query["platform"].region.value
                })

        return generator()

    _validate_get_champion_mastery_score_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(int)

    @get.register(ChampionMasteryScoreDto)
    def get_champion_mastery_score(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionMasteryScoreDto:
        ChampionMasteryAPI._validate_get_champion_mastery_score_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/champion-mastery/v3/scores/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=query["summoner.id"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "scores/by-summoner/summonerId"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        return ChampionMasteryScoreDto({
            "region": query["platform"].region.value,
            "summonerId": query["summoner.id"],
            "score": data
        })

    _validate_get_many_champion_mastery_score = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.ids").as_(Iterable)

    @get_many.register(ChampionMasteryScoreDto)
    def get_many_champion_mastery_score(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionMasteryScoreDto, None, None]:
        ChampionMasteryAPI._validate_get_many_champion_mastery_score(query, context)

        def generator():
            for summoner_id in query["summoner.ids"]:
                url = "https://{platform}.api.riotgames.com/lol/champion-mastery/v3/scores/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=summoner_id)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "scores/by-summoner/summonerId"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                yield ChampionMasteryScoreDto({
                    "region": query["platform"].region.value,
                    "summonerId": summoner_id,
                    "score": data
                })

        return generator()



from typing import Iterable, Set

from datapipelines import CompositeDataSource
from .common import RiotAPIService


def _default_services(api_key: str) -> Set[RiotAPIService]:
    from ..common import HTTPClient
    from ..image import ImageDataSource
    from .staticdata import StaticDataAPI
    from .champion import ChampionAPI
    from .summoner import SummonerAPI
    from .championmastery import ChampionMasteryAPI
    from .runepage import RunePageAPI
    from .masterypage import MasteryPageAPI
    from .match import MatchAPI
    from .spectator import SpectatorAPI
    from .status import StatusAPI
    from .leagues import LeaguesAPI

    application_rate_limiters = {}

    client = HTTPClient()
    services = {
        ImageDataSource(client),
        ChampionAPI(api_key, client, application_rate_limiters),
        StaticDataAPI(api_key, client, application_rate_limiters),
        SummonerAPI(api_key, client, application_rate_limiters),
        ChampionMasteryAPI(api_key, client, application_rate_limiters),
        RunePageAPI(api_key, client, application_rate_limiters),
        MasteryPageAPI(api_key, client, application_rate_limiters),
        MatchAPI(api_key, client, application_rate_limiters),
        SpectatorAPI(api_key, client, application_rate_limiters),
        StatusAPI(api_key, client, application_rate_limiters),
        LeaguesAPI(api_key, client, application_rate_limiters)
    }

    return services


class RiotAPI(CompositeDataSource):
    def __init__(self, api_key: str, services: Iterable[RiotAPIService] = None) -> None:
        if services is None:
            services = _default_services(api_key)

        super().__init__(services)



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Platform, Region, Queue
from ...dto.league import LeaguesListDto, ChallengerLeagueListDto, MasterLeagueListDto

T = TypeVar("T")


class LeaguesAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_leagues_query = Query. \
        has("summoner.id").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(LeaguesListDto)
    def get_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LeaguesListDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        LeaguesAPI._validate_get_leagues_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/league/v3/leagues/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=query["summoner.id"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "leagues/by-summoner/summonerId"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data = {"leagues": data}
        data["region"] = query["platform"].region.value
        data["summonerId"] = query["summoner.id"]
        return LeaguesListDto(data)

    _validate_get_many_leagues_query = Query. \
        has("summoner.ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(LeaguesListDto)
    def get_leagues(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[LeaguesListDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        LeaguesAPI._validate_get_many_leagues_query(query, context)

        def generator():
            for id in query["summoner.ids"]:
                url = "https://{platform}.api.riotgames.com/lol/league/v3/leagues/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=id)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "leagues/by-summoner/summonerId"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"leagues": data}
                data["region"] = query["platform"].region.value
                data["summonerId"] = id
                yield LeaguesListDto(data)

        return generator()


    _validate_get_challenger_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    @get.register(ChallengerLeagueListDto)
    def get_challenger_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChallengerLeagueListDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        LeaguesAPI._validate_get_challenger_league_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/league/v3/challengerleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=query["queue"].value)
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "challengerleagues/by-queue"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["queue"] = query["queue"].value
        return ChallengerLeagueListDto(data)

    _validate_get_many_challenger_league_query = Query. \
        has("queues").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(ChallengerLeagueListDto)
    def get_challenger_leagues(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChallengerLeagueListDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        LeaguesAPI._validate_get_many_master_league_query(query, context)

        def generator():
            for queue in query["queues"]:
                url = "https://{platform}.api.riotgames.com/lol/league/v3/challengerleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=queue.value)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "challengerleagues/by-queue"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"leagues": data}
                data["region"] = query["platform"].region.value
                data["queue"] = queue.value
                yield ChallengerLeagueListDto(data)

        return generator()


    _validate_get_master_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    @get.register(MasterLeagueListDto)
    def get_master_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MasterLeagueListDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        LeaguesAPI._validate_get_master_league_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/league/v3/masterleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=query["queue"].value)
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "masterleagues/by-queue"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["queue"] = query["queue"].value
        return MasterLeagueListDto(data)

    _validate_get_many_master_league_query = Query. \
        has("queues").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(MasterLeagueListDto)
    def get_master_leagues(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MasterLeagueListDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        LeaguesAPI._validate_get_many_master_league_query(query, context)

        def generator():
            for queue in query["queues"]:
                url = "https://{platform}.api.riotgames.com/lol/league/v3/masterleagues/by-queue/{queueName}".format(platform=query["platform"].value.lower(), queueName=queue.value)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "masterleagues/by-queue"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data = {"leagues": data}
                data["region"] = query["platform"].region.value
                data["queue"] = queue.value
                yield MasterLeagueListDto(data)

        return generator()



# Core types
from .champion import Champion, Champions
from .mastery import Mastery, Masteries
from .rune import Rune, Runes
from .item import Item, Items
from .summonerspell import SummonerSpell, SummonerSpells
from .map import Map, Maps
from .profileicon import ProfileIcon, ProfileIcons
from .languagestrings import LanguageStrings
from .language import Locales
from .realm import Realms
from .version import Versions



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Platform, Region
from ...dto.runepage import RunePagesDto

T = TypeVar("T")


class RunePageAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_rune_pages_query = Query. \
        has("summoner.id").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(RunePagesDto)
    def get_runes_pages(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RunePagesDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        RunePageAPI._validate_get_rune_pages_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/platform/v3/runes/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=query["summoner.id"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "runes/by-summoner/summonerId"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return RunePagesDto(data)

    _validate_get_many_rune_pages_query = Query. \
        has("summoner.ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(RunePagesDto)
    def get_rune_pages(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[RunePagesDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        RunePageAPI._validate_get_many_rune_pages_query(query, context)

        def generator():
            for id in query["summoner.ids"]:
                url = "https://{platform}.api.riotgames.com/lol/platform/v3/runes/by-summoner/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=id)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "runes/by-summoner/summonerId"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = query["platform"].region.value
                yield RunePagesDto(data)

        return generator()



import json
from typing import Type, TypeVar, MutableMapping, Any, Iterable

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from ..data import Platform
from ..dto.staticdata.champion import ChampionListDto
from ..dto.staticdata.mastery import MasteryListDto
from ..dto.staticdata.rune import RuneListDto
from ..dto.staticdata.item import ItemListDto
from ..dto.staticdata.summonerspell import SummonerSpellListDto
from ..dto.staticdata.version import VersionListDto
from ..dto.staticdata.profileicon import ProfileIconDataDto
from ..dto.staticdata.language import LanguagesDto, LanguageStringsDto
from ..dto.staticdata.realm import RealmDto
from ..dto.staticdata.map import MapListDto

from .common import HTTPClient, HTTPError
from .riotapi.staticdata import _get_default_version

T = TypeVar("T")


class DDragonDataSource(DataSource):
    def __init__(self, http_client: HTTPClient = None) -> None:
        if http_client is None:
            self._client = HTTPClient()
        else:
            self._client = http_client

    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    #############
    # Champions #
    #############

    _validate_get_champion_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").also. \
        can_have("includedData")

    @get.register(ChampionListDto)
    def get_champion_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionListDto:
        self._validate_get_champion_list_query(query, context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/championFull.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        for champ_name, champ in body["data"].items():
            champ["id"], champ["key"] = int(champ["key"]), champ["id"]

            for skin in champ["skins"]:
                # id str in DDragon, int in static data.
                skin["id"] = int(skin["id"])
                # Doesn't exist in static data.
                skin.pop("chromas", None)

            champ["passive"]["sanitizedDescription"] = champ["passive"]["description"]

            for recommended in champ['recommended']:
                # These fields always(?) are the same and don't appear in static data.
                [recommended.pop(key, None) for key in
                 ('sortrank', 'extensionPage', 'customPanel', 'customTag', 'requiredPerk', 'customPanelCurrencyType',
                  'customPanelBuffCurrencyName')]

                for block in recommended['blocks']:
                    # These don't appear in static data for whatever reason.
                    [block.pop(key, None) for key in
                     ("recSteps", "minSummonerLevel", "maxSummonerLevel", "showIfSummonerSpell", "hideIfSummonerSpell")]

                    for item in block["items"]:
                        # id str in DDragon, int in static data.
                        item["id"] = int(item["id"])
                        # Doesn't exist.
                        item.pop("hideCount", None)

            for spell in champ['spells']:
                # id -> key
                spell["key"] = spell.pop("id")
                # effectBurn is null in DDragon, empty string in static data.
                spell["effectBurn"][0] = ""
                # TODO: Sanitizer?
                spell["sanitizedDescription"] = spell["description"]
                spell["sanitizedTooltip"] = spell["tooltip"]
                # non-existent in static data(? used for charge based spells, not sure why static data strips it)
                spell.pop("maxammo", None)

                for var in spell["vars"]:
                    # coeff is always a list, even if just one item
                    if not isinstance(var["coeff"], list):
                        var["coeff"] = [var["coeff"]]

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        body["includedData"] = {"all"}
        return ChampionListDto(body)

    ############
    # Versions #
    ############

    _validate_get_versions_query = Query. \
        has("platform").as_(Platform)

    @get.register(VersionListDto)
    def get_versions(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> VersionListDto:
        self._validate_get_versions_query(query, context)

        url = "https://ddragon.leagueoflegends.com/api/versions.json"
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        return VersionListDto({
            "region": query["platform"].region.value,
            "versions": body
        })

    ##########
    # Realms #
    ##########

    _validate_get_realms_query = Query. \
        has("platform").as_(Platform)

    @get.register(RealmDto)
    def get_realms(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RealmDto:
        self._validate_get_realms_query(query, context)

        region = query["platform"].region
        url = "https://ddragon.leagueoflegends.com/realms/{region}.json".format(region=region.value.lower())
        try:
            body = json.loads(self._client.get(url)[0])

        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        body["region"] = query["platform"].region.value
        return RealmDto(body)

    #############
    # Languages #
    #############

    _validate_get_languages_query = Query. \
        has("platform").as_(Platform)

    @get.register(LanguagesDto)
    def get_languages(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LanguagesDto:
        self._validate_get_languages_query(query, context)

        url = "https://ddragon.leagueoflegends.com/cdn/languages.json"
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        data = {"region": query["platform"].region.value, "languages": body}
        return LanguagesDto(data)

    ########
    # Maps #
    ########

    _validate_get_map_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").as_(str)

    @get.register(MapListDto)
    def get_map_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MapListDto:
        self._validate_get_map_list_query(query, context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/map.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        for map in body["data"].values():
            map["mapName"] = map.pop("MapName")
            map["mapId"] = map.pop("MapId")
        return MapListDto(body)

    ####################
    # Language Strings #
    ####################

    _validate_get_language_strings_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").as_(str)

    @get.register(LanguageStringsDto)
    def get_language_strings(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LanguageStringsDto:
        self._validate_get_language_strings_query(query, context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/language.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        return LanguageStringsDto(body)

    #############
    # Masteries #
    #############

    _validate_get_mastery_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData")

    @get.register(MasteryListDto)
    def get_mastery_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MasteryListDto:
        self._validate_get_mastery_list_query(query,  context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/mastery.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        mastery_map = {}  # mastery ID -> tree name mapping. Supplied by static data but not DDragon.

        # Fix tree structure to match static data.
        for tree in body["tree"]:
            for x, tier in enumerate(body["tree"][tree]):
                for mastery in tier:
                    # In DDragon, masteryId is a str whereas in static data it is an int.
                    mastery_id = int(mastery["masteryId"])
                    mastery["masteryId"] = mastery_id
                    mastery_map[mastery_id] = tree
                # Each tier in the static data mastery tree is encapsulated by a single item dictionary.
                body["tree"][tree][x] = {"masteryTreeItems": tier}

        # Add fields not provided by DDragon
        for mastery in body["data"].values():
            mastery["masteryTree"] = mastery_map[mastery["id"]]
            # TODO: Sanitizer?
            mastery["sanitizedDescription"] = mastery["description"]

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        body["includedData"] = {"all"}
        return MasteryListDto(body)

    #########
    # Runes #
    #########

    _validate_get_rune_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData")

    @get.register(RuneListDto)
    def get_rune_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RuneListDto:
        self._validate_get_rune_list_query(query, context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/rune.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        body.pop("basic")

        for rune_id, rune in body["data"].items():
            rune["id"] = int(rune_id)
            # TODO: Sanitizer?
            rune["sanitizedDescription"] = rune["description"]
            # Some DDragon stats begin with "r", but they don"t in static data.
            rune["stats"] = {k.strip("r"): v for k, v in rune["stats"].items()}
            # isrune on DDragon is isRune on static data.
            rune["rune"]["isRune"] = rune["rune"].pop("isrune")
            # colloq and plaintext are always(?) null, and don"t appear in static data.
            rune.pop("colloq")
            rune.pop("plaintext")

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        body["includedData"] = {"all"}
        return RuneListDto(body)

    #########
    # Items #
    #########

    _validate_get_item_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData")

    @get.register(ItemListDto)
    def get_item_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ItemListDto:
        self._validate_get_item_list_query(query, context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/item.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        body.pop("basic")

        for group in body["groups"]:
            # key in static data -> id on DDragon
            group["key"] = group.pop("id")

        for item_id, item in body["data"].items():
            item["id"] = int(item_id)
            # TODO: Sanitizer?
            item["sanitizedDescription"] = item["description"]

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        body["includedData"] = {"all"}

        return ItemListDto(body)

    ###################
    # Summoner Spells #
    ###################

    _validate_get_summoner_spell_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").as_(str).also. \
        can_have("includedData")

    @get.register(SummonerSpellListDto)
    def get_summoner_spell_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerSpellListDto:
        self._validate_get_summoner_spell_list_query(query, context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/summoner.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        for ss_name, ss in body["data"].items():
            # key and id are switched between DDragon and static data. Also, id is of type int, instead of str.
            ss["id"], ss["key"] = int(ss["key"]), ss["id"]
            # effectBurn"s first element is an null in DDragon, but an empty string in static data..
            ss["effectBurn"][0] = ""
            # Usually -1, doesn"t exist in static data.
            ss.pop("maxammo")
            # TODO: Sanitizer?
            ss["sanitizedDescription"] = ss["description"]
            ss["sanitizedTooltip"] = ss["tooltip"]

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        body["includedData"] = {"all"}
        return SummonerSpellListDto(body)

    #################
    # Profile Icons #
    #################

    _validate_get_profile_icon_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_default_version, supplies_type=str).also. \
        can_have("locale").as_(str)

    @get.register(ProfileIconDataDto)
    def get_profile_icon(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ProfileIconDataDto:
        self._validate_get_profile_icon_query(query, context)

        locale = query["locale"] if "locale" in query else query["platform"].default_locale

        url = "http://ddragon.leagueoflegends.com/cdn/{version}/data/{locale}/profileicon.json".format(
            version=query["version"],
            locale=locale
        )
        try:
            body = json.loads(self._client.get(url)[0])
        except HTTPError as e:
            raise NotFoundError(str(e)) from e

        body["region"] = query["platform"].region.value
        body["locale"] = locale
        return ProfileIconDataDto(body)



#!/usr/bin/env python

import sys

from setuptools import setup, find_packages


install_requires = [
    "datapipelines",
    "merakicommons",
    "Pillow",
    "pycurl"
]

# Require python 3.6
if sys.version_info.major != 3 and sys.version_info.minor != 6:
    sys.exit("Cassiopeia requires Python 3.6.")

setup(
    name="cassiopeia",
    version="3.0.7-rc1",
    author="Jason Maldonis; Rob Rua",
    author_email="team@merakianalytics.com",
    url="https://github.com/meraki-analytics/cassiopeia",
    description="Riot Games Developer API Wrapper (3rd Party)",
    keywords=["LoL", "League of Legends", "Riot Games", "API", "REST"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Games/Entertainment",
        "Topic :: Games/Entertainment :: Real Time Strategy",
        "Topic :: Games/Entertainment :: Role-Playing",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    license="MIT",
    packages=find_packages(),
    zip_safe=True,
    install_requires=install_requires,
    include_package_data=True
)


from typing import List, Set, Union
from PIL.Image import Image as PILImage

from merakicommons.ghost import ghost_load_on
from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList

from ...configuration import settings
from ...data import Region, Platform, RuneType
from ..common import CoreData, DataObjectList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaGhostList, get_latest_version
from .common import ImageData, Sprite, Image
from ...dto.staticdata import rune as dto


##############
# Data Types #
##############


class RuneListData(DataObjectList):
    _dto_type = dto.RuneListDto
    _renamed = {"included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]


class MetadataData(CoreData):
    _renamed = {"is_rune": "isRune"}

    # TODO Convert the type in the transformer from str to int
    @property
    def tier(self) -> int:
        return int(self._dto["tier"])

    @property
    def type(self) -> str:
        return self._dto["type"]

    @property
    def is_rune(self) -> bool:
        return self._dto["is_rune"]


class RuneStatsData(CoreData):
    _renamed = {"percent_time_dead_mod_per_level": "PercentTimeDeadModPerLevel", "percent_armor_penetration_mod_per_level": "PercentArmorPenetrationModPerLevel", "percent_crit_damage_mod": "PercentCritDamageMod", "percent_spell_block_mod": "PercentSpellBlockMod", "percent_hp_regen_mod": "PercentHPRegenMod", "percent_movement_speed_mod": "PercentMovementSpeedMod", "flat_spell_block_mod": "FlatSpellBlockMod", "flat_energy_regen_mod_per_level": "FlatEnergyRegenModPerLevel", "flat_energy_pool_mod": "FlatEnergyPoolMod", "flat_magic_penetration_mod_per_level": "FlatMagicPenetrationModPerLevel", "percent_life_steal_mod": "PercentLifeStealMod", "flat_mp_pool_mod": "FlatMPPoolMod", "percent_cooldown_mod": "PercentCooldownMod", "percent_magic_penetration_mod": "PercentMagicPenetrationMod", "flat_armor_penetration_mod_per_level": "FlatArmorPenetrationModPerLevel", "flat_movement_speed_mod": "FlatMovementSpeedMod", "flat_time_dead_mod_per_level": "FlatTimeDeadModPerLevel", "flat_armor_mod_per_level": "FlatArmorModPerLevel", "percent_attack_speed_mod": "PercentAttackSpeedMod", "flat_dodge_mod_per_level": "FlatDodgeModPerLevel", "percent_magic_damage_mod": "PercentMagicDamageMod", "percent_block_mod": "PercentBlockMod", "flat_dodge_mod": "FlatDodgeMod", "flat_energy_regen_mod": "FlatEnergyRegenMod", "flat_hp_mod_per_level": "FlatHPModPerLevel", "percent_attack_speed_mod_per_level": "PercentAttackSpeedModPerLevel", "percent_spell_vamp_mod": "PercentSpellVampMod", "flat_mp_regen_mod": "FlatMPRegenMod", "percent_hp_pool_mod": "PercentHPPoolMod", "percent_dodge_mod": "PercentDodgeMod", "flat_attack_speed_mod": "FlatAttackSpeedMod", "flat_armor_mod": "FlatArmorMod", "flat_magic_damage_mod_per_level": "FlatMagicDamageModPerLevel", "flat_hp_regen_mod": "FlatHPRegenMod", "percent_physical_damage_mod": "PercentPhysicalDamageMod", "flat_crit_chance_mod_per_level": "FlatCritChanceModPerLevel", "flat_spell_block_mod_per_level": "FlatSpellBlockModPerLevel", "percent_time_dead_mod": "PercentTimeDeadMod", "flat_block_mod": "FlatBlockMod", "percent_mp_pool_mod": "PercentMPPoolMod", "flat_magic_damage_mod": "FlatMagicDamageMod", "percent_mp_regen_mod": "PercentMPRegenMod", "percent_movement_speed_mod_per_level": "PercentMovementSpeedModPerLevel", "percent_cooldown_mod_per_level": "PercentCooldownModPerLevel", "flat_mp_mod_per_level": "FlatMPModPerLevel", "flat_energy_mod_per_level": "FlatEnergyModPerLevel", "flat_physical_damage_mod": "FlatPhysicalDamageMod", "flat_hp_regen_mod_per_level": "FlatHPRegenModPerLevel", "flat_crit_damage_mod": "FlatCritDamageMod", "percent_armor_mod": "PercentArmorMod", "flat_magic_penetration_mod": "FlatMagicPenetrationMod", "percent_crit_chance_mod": "PercentCritChanceMod", "flat_physical_damage_mod_per_level": "FlatPhysicalDamageModPerLevel", "percent_armor_penetration_mod": "PercentArmorPenetrationMod", "percent_exp_bonus": "PercentEXPBonus", "flat_mp_regen_mod_per_level": "FlatMPRegenModPerLevel", "percent_magic_penetration_mod_per_level": "PercentMagicPenetrationModPerLevel", "flat_time_dead_mod": "FlatTimeDeadMod", "flat_movement_speed_mod_per_level": "FlatMovementSpeedModPerLevel", "flat_gold_per_10_mod": "FlatGoldPer10Mod", "flat_armor_penetration_mod": "FlatArmorPenetrationMod", "flat_crit_damage_mod_per_level": "FlatCritDamageModPerLevel", "flat_hp_pool_mod": "FlatHPPoolMod", "flat_crit_chance_mod": "FlatCritChanceMod", "flat_exp_bonus": "FlatEXPBonus"}

    @property
    def percent_time_dead_mod_per_level(self) -> float:
        return self._dto["PercentTimeDeadModPerLevel"]

    @property
    def percent_armor_penetration_mod_per_level(self) -> float:
        return self._dto["PercentArmorPenetrationModPerLevel"]

    @property
    def percent_crit_damage_mod(self) -> float:
        return self._dto["PercentCritDamageMod"]

    @property
    def percent_spell_block_mod(self) -> float:
        return self._dto["PercentSpellBlockMod"]

    @property
    def percent_hp_regen_mod(self) -> float:
        return self._dto["PercentHPRegenMod"]

    @property
    def percent_movement_speed_mod(self) -> float:
        return self._dto["PercentMovementSpeedMod"]

    @property
    def flat_spell_block_mod(self) -> float:
        return self._dto["FlatSpellBlockMod"]

    @property
    def flat_energy_regen_mod_per_level(self) -> float:
        return self._dto["FlatEnergyRegenModPerLevel"]

    @property
    def flat_energy_pool_mod(self) -> float:
        return self._dto["FlatEnergyPoolMod"]

    @property
    def flat_magic_penetration_mod_per_level(self) -> float:
        return self._dto["FlatMagicPenetrationModPerLevel"]

    @property
    def percent_life_steal_mod(self) -> float:
        return self._dto["PercentLifeStealMod"]

    @property
    def flat_mp_pool_mod(self) -> float:
        return self._dto["FlatMPPoolMod"]

    @property
    def percent_cooldown_mod(self) -> float:
        return self._dto["PercentCooldownMod"]

    @property
    def percent_magic_penetration_mod(self) -> float:
        return self._dto["PercentMagicPenetrationMod"]

    @property
    def flat_armor_penetration_mod_per_level(self) -> float:
        return self._dto["FlatArmorPenetrationModPerLevel"]

    @property
    def flat_movement_speed_mod(self) -> float:
        return self._dto["FlatMovementSpeedMod"]

    @property
    def flat_time_dead_mod_per_level(self) -> float:
        return self._dto["FlatTimeDeadModPerLevel"]

    @property
    def flat_armor_mod_per_level(self) -> float:
        return self._dto["FlatArmorModPerLevel"]

    @property
    def percent_attack_speed_mod(self) -> float:
        return self._dto["PercentAttackSpeedMod"]

    @property
    def flat_dodge_mod_per_level(self) -> float:
        return self._dto["FlatDodgeModPerLevel"]

    @property
    def percent_magic_damage_mod(self) -> float:
        return self._dto["PercentMagicDamageMod"]

    @property
    def percent_block_mod(self) -> float:
        return self._dto["PercentBlockMod"]

    @property
    def flat_dodge_mod(self) -> float:
        return self._dto["FlatDodgeMod"]

    @property
    def flat_energy_regen_mod(self) -> float:
        return self._dto["FlatEnergyRegenMod"]

    @property
    def flat_hp_mod_per_level(self) -> float:
        return self._dto["FlatHPModPerLevel"]

    @property
    def percent_attack_speed_mod_per_level(self) -> float:
        return self._dto["PercentAttackSpeedModPerLevel"]

    @property
    def percent_spell_vamp_mod(self) -> float:
        return self._dto["PercentSpellVampMod"]

    @property
    def flat_mp_regen_mod(self) -> float:
        return self._dto["FlatMPRegenMod"]

    @property
    def percent_hp_pool_mod(self) -> float:
        return self._dto["PercentHPPoolMod"]

    @property
    def percent_dodge_mod(self) -> float:
        return self._dto["PercentDodgeMod"]

    @property
    def flat_attack_speed_mod(self) -> float:
        return self._dto["FlatAttackSpeedMod"]

    @property
    def flat_armor_mod(self) -> float:
        return self._dto["FlatArmorMod"]

    @property
    def flat_magic_damage_mod_per_level(self) -> float:
        return self._dto["FlatMagicDamageModPerLevel"]

    @property
    def flat_hp_regen_mod(self) -> float:
        return self._dto["FlatHPRegenMod"]

    @property
    def percent_physical_damage_mod(self) -> float:
        return self._dto["PercentPhysicalDamageMod"]

    @property
    def flat_crit_chance_mod_per_level(self) -> float:
        return self._dto["FlatCritChanceModPerLevel"]

    @property
    def flat_spell_block_mod_per_level(self) -> float:
        return self._dto["FlatSpellBlockModPerLevel"]

    @property
    def percent_time_dead_mod(self) -> float:
        return self._dto["PercentTimeDeadMod"]

    @property
    def flat_block_mod(self) -> float:
        return self._dto["FlatBlockMod"]

    @property
    def percent_mp_pool_mod(self) -> float:
        return self._dto["PercentMPPoolMod"]

    @property
    def flat_magic_damage_mod(self) -> float:
        return self._dto["FlatMagicDamageMod"]

    @property
    def percent_mp_regen_mod(self) -> float:
        return self._dto["PercentMPRegenMod"]

    @property
    def percent_movement_speed_mod_per_level(self) -> float:
        return self._dto["PercentMovementSpeedModPerLevel"]

    @property
    def percent_cooldown_mod_per_level(self) -> float:
        return self._dto["PercentCooldownModPerLevel"]

    @property
    def flat_mp_mod_per_level(self) -> float:
        return self._dto["FlatMPModPerLevel"]

    @property
    def flat_energy_mod_per_level(self) -> float:
        return self._dto["FlatEnergyModPerLevel"]

    @property
    def flat_physical_damage_mod(self) -> float:
        return self._dto["FlatPhysicalDamageMod"]

    @property
    def flat_hp_regen_mod_per_level(self) -> float:
        return self._dto["FlatHPRegenModPerLevel"]

    @property
    def flat_crit_damage_mod(self) -> float:
        return self._dto["FlatCritDamageMod"]

    @property
    def percent_armor_mod(self) -> float:
        return self._dto["PercentArmorMod"]

    @property
    def flat_magic_penetration_mod(self) -> float:
        return self._dto["FlatMagicPenetrationMod"]

    @property
    def percent_crit_chance_mod(self) -> float:
        return self._dto["PercentCritChanceMod"]

    @property
    def flat_physical_damage_mod_per_level(self) -> float:
        return self._dto["FlatPhysicalDamageModPerLevel"]

    @property
    def percent_armor_penetration_mod(self) -> float:
        return self._dto["PercentArmorPenetrationMod"]

    @property
    def percent_exp_bonus(self) -> float:
        return self._dto["PercentEXPBonus"]

    @property
    def flat_mp_regen_mod_per_level(self) -> float:
        return self._dto["FlatMPRegenModPerLevel"]

    @property
    def percent_magic_penetration_mod_per_level(self) -> float:
        return self._dto["PercentMagicPenetrationModPerLevel"]

    @property
    def flat_time_dead_mod(self) -> float:
        return self._dto["FlatTimeDeadMod"]

    @property
    def flat_movement_speed_mod_per_level(self) -> float:
        return self._dto["FlatMovementSpeedModPerLevel"]

    @property
    def flat_gold_per_10_mod(self) -> float:
        return self._dto["FlatGoldPer10Mod"]

    @property
    def flat_armor_penetration_mod(self) -> float:
        return self._dto["FlatArmorPenetrationMod"]

    @property
    def flat_crit_damage_mod_per_level(self) -> float:
        return self._dto["FlatCritDamageModPerLevel"]

    @property
    def flat_hp_pool_mod(self) -> float:
        return self._dto["FlatHPPoolMod"]

    @property
    def flat_crit_chance_mod(self) -> float:
        return self._dto["FlatCritChanceMod"]

    @property
    def flat_exp_bonus(self) -> float:
        return self._dto["FlatEXPBonus"]


class RuneData(CoreData):
    _dto_type = dto.RuneDto
    _renamed = {"metadata": "rune", "sanitized_description": "sanitizedDescription", "included_data": "includedData"}

    @property
    def region(self) -> str:
        return self._dto["region"]

    @property
    def version(self) -> str:
        return self._dto["version"]

    @property
    def locale(self) -> str:
        return self._dto["locale"]

    @property
    def included_data(self) -> Set[str]:
        return self._dto["includedData"]

    @property
    def metadata(self) -> MetadataData:
        return MetadataData.from_dto(self._dto["rune"])

    @property
    def name(self) -> str:
        return self._dto["name"]

    @property
    def id(self) -> int:
        return self._dto["id"]

    @property
    def stats(self) -> RuneStatsData:
        return RuneStatsData.from_dto(self._dto["stats"])

    @property
    def image(self) -> ImageData:
        return ImageData.from_dto(self._dto["image"])

    @property
    def tags(self) -> List[str]:
        return self._dto["tags"]

    @property
    def sanitized_description(self) -> str:
        return self._dto["sanitizedDescription"]

    @property
    def description(self) -> str:
        return self._dto["description"]


##############
# Core Types #
##############


class Runes(CassiopeiaGhostList):
    _data_types = {RuneListData}

    def __init__(self, *args, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        super().__init__(*args, **kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        self.clear()
        from ...transformers.staticdata import StaticDataTransformer
        SearchableList.__init__(self, [StaticDataTransformer.rune_data_to_core(None, i) for i in data])
        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[RuneListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[RuneListData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="rune")
            self(version=version)
            return self._data[RuneListData].version

    @property
    def locale(self) -> str:
        """The locale for this champion."""
        return self._data[RuneListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this champion when it's loaded."""
        return self._data[RuneListData].included_data


class RuneStats(CassiopeiaObject):
    _data_types = {RuneStatsData}

    @property
    def percent_time_dead_mod_per_level(self) -> float:
        return self._data[RuneStatsData].percent_time_dead_mod_per_level

    @property
    def percent_armor_penetration_mod_per_level(self) -> float:
        return self._data[RuneStatsData].percent_armor_penetration_mod_per_level

    @property
    def percent_crit_damage_mod(self) -> float:
        return self._data[RuneStatsData].percent_crit_damage_mod

    @property
    def percent_spell_block_mod(self) -> float:
        return self._data[RuneStatsData].percent_spell_block_mod

    @property
    def percent_hp_regen_mod(self) -> float:
        return self._data[RuneStatsData].percent_hp_regen_mod

    @property
    def percent_movement_speed_mod(self) -> float:
        return self._data[RuneStatsData].percent_movement_speed_mod

    @property
    def flat_spell_block_mod(self) -> float:
        return self._data[RuneStatsData].flat_spell_block_mod

    @property
    def flat_energy_regen_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_energy_regen_mod_per_level

    @property
    def flat_energy_pool_mod(self) -> float:
        return self._data[RuneStatsData].flat_energy_pool_mod

    @property
    def flat_magic_penetration_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_magic_penetration_mod_per_level

    @property
    def percent_life_steal_mod(self) -> float:
        return self._data[RuneStatsData].percent_life_steal_mod

    @property
    def flat_mp_pool_mod(self) -> float:
        return self._data[RuneStatsData].flat_mp_pool_mod

    @property
    def percent_cooldown_mod(self) -> float:
        return self._data[RuneStatsData].percent_cooldown_mod

    @property
    def percent_magic_penetration_mod(self) -> float:
        return self._data[RuneStatsData].percent_magic_penetration_mod

    @property
    def flat_armor_penetration_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_armor_penetration_mod_per_level

    @property
    def flat_movement_speed_mod(self) -> float:
        return self._data[RuneStatsData].flat_movement_speed_mod

    @property
    def flat_time_dead_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_time_dead_mod_per_level

    @property
    def flat_armor_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_armor_mod_per_level

    @property
    def percent_attack_speed_mod(self) -> float:
        return self._data[RuneStatsData].percent_attack_speed_mod

    @property
    def flat_dodge_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_dodge_mod_per_level

    @property
    def percent_magic_damage_mod(self) -> float:
        return self._data[RuneStatsData].percent_magic_damage_mod

    @property
    def percent_block_mod(self) -> float:
        return self._data[RuneStatsData].percent_block_mod

    @property
    def flat_dodge_mod(self) -> float:
        return self._data[RuneStatsData].flat_dodge_mod

    @property
    def flat_energy_regen_mod(self) -> float:
        return self._data[RuneStatsData].flat_energy_regen_mod

    @property
    def flat_hp_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_hp_mod_per_level

    @property
    def percent_attack_speed_mod_per_level(self) -> float:
        return self._data[RuneStatsData].percent_attack_speed_mod_per_level

    @property
    def percent_spell_vamp_mod(self) -> float:
        return self._data[RuneStatsData].percent_spell_vamp_mod

    @property
    def flat_mp_regen_mod(self) -> float:
        return self._data[RuneStatsData].flat_mp_regen_mod

    @property
    def percent_hp_pool_mod(self) -> float:
        return self._data[RuneStatsData].percent_hp_pool_mod

    @property
    def percent_dodge_mod(self) -> float:
        return self._data[RuneStatsData].percent_dodge_mod

    @property
    def flat_attack_speed_mod(self) -> float:
        return self._data[RuneStatsData].flat_attack_speed_mod

    @property
    def flat_armor_mod(self) -> float:
        return self._data[RuneStatsData].flat_armor_mod

    @property
    def flat_magic_damage_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_magic_damage_mod_per_level

    @property
    def flat_hp_regen_mod(self) -> float:
        return self._data[RuneStatsData].flat_hp_regen_mod

    @property
    def percent_physical_damage_mod(self) -> float:
        return self._data[RuneStatsData].percent_physical_damage_mod

    @property
    def flat_crit_chance_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_crit_chance_mod_per_level

    @property
    def flat_spell_block_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_spell_block_mod_per_level

    @property
    def percent_time_dead_mod(self) -> float:
        return self._data[RuneStatsData].percent_time_dead_mod

    @property
    def flat_block_mod(self) -> float:
        return self._data[RuneStatsData].flat_block_mod

    @property
    def percent_mp_pool_mod(self) -> float:
        return self._data[RuneStatsData].percent_mp_pool_mod

    @property
    def flat_magic_damage_mod(self) -> float:
        return self._data[RuneStatsData].flat_magic_damage_mod

    @property
    def percent_mp_regen_mod(self) -> float:
        return self._data[RuneStatsData].percent_mp_regen_mod

    @property
    def percent_movement_speed_mod_per_level(self) -> float:
        return self._data[RuneStatsData].percent_movement_speed_mod_per_level

    @property
    def percent_cooldown_mod_per_level(self) -> float:
        return self._data[RuneStatsData].percent_cooldown_mod_per_level

    @property
    def flat_mp_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_mp_mod_per_level

    @property
    def flat_energy_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_energy_mod_per_level

    @property
    def flat_physical_damage_mod(self) -> float:
        return self._data[RuneStatsData].flat_physical_damage_mod

    @property
    def flat_hp_regen_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_hp_regen_mod_per_level

    @property
    def flat_crit_damage_mod(self) -> float:
        return self._data[RuneStatsData].flat_crit_damage_mod

    @property
    def percent_armor_mod(self) -> float:
        return self._data[RuneStatsData].percent_armor_mod

    @property
    def flat_magic_penetration_mod(self) -> float:
        return self._data[RuneStatsData].flat_magic_penetration_mod

    @property
    def percent_crit_chance_mod(self) -> float:
        return self._data[RuneStatsData].percent_crit_chance_mod

    @property
    def flat_physical_damage_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_physical_damage_mod_per_level

    @property
    def percent_armor_penetration_mod(self) -> float:
        return self._data[RuneStatsData].percent_armor_penetration_mod

    @property
    def percent_exp_bonus(self) -> float:
        return self._data[RuneStatsData].percent_exp_bonus

    @property
    def flat_mp_regen_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_mp_regen_mod_per_level

    @property
    def percent_magic_penetration_mod_per_level(self) -> float:
        return self._data[RuneStatsData].percent_magic_penetration_mod_per_level

    @property
    def flat_time_dead_mod(self) -> float:
        return self._data[RuneStatsData].flat_time_dead_mod

    @property
    def flat_movement_speed_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_movement_speed_mod_per_level

    @property
    def flat_gold_per_10_mod(self) -> float:
        return self._data[RuneStatsData].flat_gold_per_10_mod

    @property
    def flat_armor_penetration_mod(self) -> float:
        return self._data[RuneStatsData].flat_armor_penetration_mod

    @property
    def flat_crit_damage_mod_per_level(self) -> float:
        return self._data[RuneStatsData].flat_crit_damage_mod_per_level

    @property
    def flat_hp_pool_mod(self) -> float:
        return self._data[RuneStatsData].flat_hp_pool_mod

    @property
    def flat_crit_chance_mod(self) -> float:
        return self._data[RuneStatsData].flat_crit_chance_mod

    @property
    def flat_exp_bonus(self) -> float:
        return self._data[RuneStatsData].flat_exp_bonus


@searchable({str: ["name", "tags", "type", "region", "platform", "locale"], int: ["id"], RuneType: ["type"], Region: ["region"], Platform: ["platform"]})
class Rune(CassiopeiaGhost):
    _data_types = {RuneData}
    _load_types = {RuneData: RuneListData}

    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if region is None:
            region = settings.default_region
        if not isinstance(region, Region):
            region = Region(region)
        if included_data is None:
            included_data = {"all"}
        if locale is None:
            locale = region.default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        def find_matching_attribute(datalist, attrname, attrvalue):
            for item in datalist:
                if getattr(item, attrname, None) == attrvalue:
                    return item

        # The `data` is a dict of summoner spell data instances
        if "id" in self._data[RuneData]._dto:
            find = "id", self.id
        elif "name" in self._data[RuneData]._dto:
            find = "name", self.name
        else:
            raise RuntimeError("Expected fields not present after loading.")
        data = find_matching_attribute(data, *find)

        super().__load_hook__(load_group, data)

    @lazy_property
    def region(self) -> Region:
        """The region for this rune."""
        return Region(self._data[RuneData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this rune."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this rune."""
        try:
            return self._data[RuneData].version
        except KeyError:
            version = get_latest_version(region=self.region, endpoint="rune")
            self(version=version)
            return self._data[RuneData].version

    @property
    def locale(self) -> str:
        """The locale for this rune."""
        return self._data[RuneData].locale

    @lazy_property
    def included_data(self) -> Set[str]:
        """The region for this rune."""
        return self._data[RuneData].included_data

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def tier(self) -> int:
        return self._data[RuneData].metadata.tier

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def type(self) -> RuneType:
        return RuneType(self._data[RuneData].metadata.type)

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def name(self) -> str:
        """The rune's name."""
        return self._data[RuneData].name

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def id(self) -> int:
        """The rune's ID."""
        return self._data[RuneData].id

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def description(self) -> int:
        """The rune's description."""
        return self._data[RuneData].description

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def sanitized_description(self) -> int:
        """The rune's sanitized description."""
        return self._data[RuneData].sanitized_description

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def stats(self) -> RuneStats:
        return RuneStats.from_data(self._data[RuneData].stats)

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    def tags(self) -> List[str]:
        return self._data[RuneData].id

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on(KeyError)
    @lazy
    def image(self) -> Image:
        """The image information for this rune."""
        image = Image(self._data[RuneData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info



import datetime

from typing import Optional as _Optional


class Patch(object):
    def __init__(self, name: str, patch: str, start: datetime.date, end: _Optional[datetime.date]):
        self._name = name
        self._patch = patch
        self._start = start
        self._end = end

    @classmethod
    def from_str(cls, string):
        for patch in cls.__patches:
            if string in patch.patch:
                return patch
        else:
            raise ValueError("Unknown patch name {}".format(string))

    @property
    def name(self):
        return self._name

    @property
    def patch(self):
        return self._patch

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def major(self):
        return self.patch.split(".")[0]

    @property
    def minor(self):
        return self.patch.split(".")[1]

    @property
    def majorminor(self):
        return ".".join(self.patch.split(".")[:2])

    @property
    def revision(self):
        return ".".join(self.patch.split(".")[2:])


patches = [
    Patch(name="Closed Beta", patch="0.9.22.4", start=datetime.date(2009, 7, 30), end=datetime.date(2009, 8, 7)),
    Patch(name="Closed Beta", patch="0.9.22.7", start=datetime.date(2009, 8, 7), end=datetime.date(2009, 8, 12)),
    Patch(name="Closed Beta", patch="0.9.22.9", start=datetime.date(2009, 8, 12), end=datetime.date(2009, 8, 19)),
    Patch(name="Closed Beta", patch="0.9.22.15", start=datetime.date(2009, 8, 19), end=datetime.date(2009, 9, 2)),
    Patch(name="Closed Beta", patch="0.9.22.16", start=datetime.date(2009, 9, 2), end=datetime.date(2009, 9, 10)),
    Patch(name="Closed Beta", patch="0.9.22.18", start=datetime.date(2009, 9, 10), end=datetime.date(2009, 9, 19)),
    Patch(name="Closed Beta", patch="0.9.25.21", start=datetime.date(2009, 9, 19), end=datetime.date(2009, 10, 1)),
    Patch(name="Closed Beta", patch="0.9.25.24", start=datetime.date(2009, 10, 1), end=datetime.date(2009, 10, 10)),
    Patch(name="Closed Beta", patch="0.9.25.34", start=datetime.date(2009, 10, 10), end=datetime.date(2009, 10, 21)),
    Patch(name="Season 1", patch="1.0.0.32", start=datetime.date(2009, 10, 21), end=datetime.date(2009, 11, 11)),
    Patch(name="Season 1", patch="1.0.0.52", start=datetime.date(2009, 11, 11), end=datetime.date(2009, 11, 20)),
    Patch(name="Season 1", patch="1.0.0.58", start=datetime.date(2009, 11, 20), end=datetime.date(2009, 12, 2)),
    Patch(name="Season 1", patch="1.0.0.61", start=datetime.date(2009, 12, 2), end=datetime.date(2009, 12, 17)),
    Patch(name="Season 1", patch="1.0.0.63", start=datetime.date(2009, 12, 17), end=datetime.date(2010, 1, 13)),
    Patch(name="Season 1", patch="1.0.0.70", start=datetime.date(2010, 1, 13), end=datetime.date(2010, 2, 2)),
    Patch(name="Season 1", patch="1.0.0.72", start=datetime.date(2010, 2, 2), end=datetime.date(2010, 2, 9)),
    Patch(name="Season 1", patch="1.0.0.74", start=datetime.date(2010, 2, 9), end=datetime.date(2010, 2, 24)),
    Patch(name="Season 1", patch="1.0.0.75", start=datetime.date(2010, 2, 24), end=datetime.date(2010, 3, 16)),
    Patch(name="Season 1", patch="1.0.0.79", start=datetime.date(2010, 3, 16), end=datetime.date(2010, 3, 24)),
    Patch(name="Season 1", patch="1.0.0.81", start=datetime.date(2010, 3, 24), end=datetime.date(2010, 4, 8)),
    Patch(name="Season 1", patch="1.0.0.82", start=datetime.date(2010, 4, 8), end=datetime.date(2010, 4, 27)),
    Patch(name="Season 1", patch="1.0.0.83", start=datetime.date(2010, 4, 27), end=datetime.date(2010, 5, 11)),
    Patch(name="Season 1", patch="1.0.0.85", start=datetime.date(2010, 5, 11), end=datetime.date(2010, 6, 1)),
    Patch(name="Season 1", patch="1.0.0.86", start=datetime.date(2010, 6, 1), end=datetime.date(2010, 6, 8)),
    Patch(name="Season 1", patch="1.0.0.87", start=datetime.date(2010, 6, 8), end=datetime.date(2010, 6, 24)),
    Patch(name="Season 1", patch="1.0.0.94", start=datetime.date(2010, 6, 24), end=datetime.date(2010, 6, 29)),
    Patch(name="Season 1", patch="1.0.0.94(b)", start=datetime.date(2010, 6, 29), end=datetime.date(2010, 7, 13)),
    Patch(name="Season 1", patch="1.0.0.96", start=datetime.date(2010, 7, 13), end=datetime.date(2010, 7, 27)),
    Patch(name="Season 1", patch="1.0.0.97", start=datetime.date(2010, 7, 27), end=datetime.date(2010, 8, 10)),
    Patch(name="Season 1", patch="1.0.0.98", start=datetime.date(2010, 8, 10), end=datetime.date(2010, 8, 24)),
    Patch(name="Season 1", patch="1.0.0.99", start=datetime.date(2010, 8, 24), end=datetime.date(2010, 9, 8)),
    Patch(name="Season 1", patch="1.0.0.100", start=datetime.date(2010, 9, 8), end=datetime.date(2010, 9, 21)),
    Patch(name="Season 1", patch="1.0.0.101", start=datetime.date(2010, 9, 21), end=datetime.date(2010, 10, 5)),
    Patch(name="Season 1", patch="1.0.0.102", start=datetime.date(2010, 10, 5), end=datetime.date(2010, 10, 19)),
    Patch(name="Season 1", patch="1.0.0.103", start=datetime.date(2010, 10, 19), end=datetime.date(2010, 11, 2)),
    Patch(name="Season 1", patch="1.0.0.104", start=datetime.date(2010, 11, 2), end=datetime.date(2010, 11, 16)),
    Patch(name="Season 1", patch="1.0.0.105", start=datetime.date(2010, 11, 16), end=datetime.date(2010, 12, 1)),
    Patch(name="Season 1", patch="1.0.0.106", start=datetime.date(2010, 12, 1), end=datetime.date(2010, 12, 14)),
    Patch(name="Season 1", patch="1.0.0.107", start=datetime.date(2010, 12, 14), end=datetime.date(2011, 1, 4)),
    Patch(name="Season 1", patch="1.0.0.108", start=datetime.date(2011, 1, 4), end=datetime.date(2011, 1, 18)),
    Patch(name="Season 1", patch="1.0.0.109", start=datetime.date(2011, 1, 18), end=datetime.date(2011, 2, 1)),
    Patch(name="Season 1", patch="1.0.0.110", start=datetime.date(2011, 2, 1), end=datetime.date(2011, 2, 16)),
    Patch(name="Season 1", patch="1.0.0.111", start=datetime.date(2011, 2, 16), end=datetime.date(2011, 3, 1)),
    Patch(name="Season 1", patch="1.0.0.112", start=datetime.date(2011, 3, 1), end=datetime.date(2011, 3, 15)),
    Patch(name="Season 1", patch="1.0.0.113", start=datetime.date(2011, 3, 15), end=datetime.date(2011, 3, 29)),
    Patch(name="Season 1", patch="1.0.0.114", start=datetime.date(2011, 3, 29), end=datetime.date(2011, 4, 12)),
    Patch(name="Season 1", patch="1.0.0.115", start=datetime.date(2011, 4, 12), end=datetime.date(2011, 4, 26)),
    Patch(name="Season 1", patch="1.0.0.116", start=datetime.date(2011, 4, 26), end=datetime.date(2011, 5, 10)),
    Patch(name="Season 1", patch="1.0.0.118", start=datetime.date(2011, 5, 10), end=datetime.date(2011, 5, 24)),
    Patch(name="Season 1", patch="1.0.0.118b", start=datetime.date(2011, 5, 24), end=datetime.date(2011, 6, 1)),
    Patch(name="Season 1", patch="1.0.0.119", start=datetime.date(2011, 6, 1), end=datetime.date(2011, 6, 22)),
    Patch(name="Season 1", patch="1.0.0.120", start=datetime.date(2011, 6, 22), end=datetime.date(2011, 7, 8)),
    Patch(name="Season 1", patch="1.0.0.121", start=datetime.date(2011, 7, 8), end=datetime.date(2011, 7, 26)),
    Patch(name="Season 1", patch="1.0.0.122", start=datetime.date(2011, 7, 26), end=datetime.date(2011, 8, 9)),
    Patch(name="Season 1", patch="1.0.0.123", start=datetime.date(2011, 8, 9), end=datetime.date(2011, 8, 24)),
    Patch(name="Season 1", patch="1.0.0.124", start=datetime.date(2011, 8, 24), end=datetime.date(2011, 9, 14)),
    Patch(name="Season 1", patch="1.0.0.125", start=datetime.date(2011, 9, 14), end=datetime.date(2011, 10, 5)),
    Patch(name="Season 2", patch="1.0.0.126", start=datetime.date(2011, 10, 5), end=datetime.date(2011, 10, 19)),
    Patch(name="Season 2", patch="1.0.0.127", start=datetime.date(2011, 10, 19), end=datetime.date(2011, 11, 1)),
    Patch(name="Season 2", patch="1.0.0.128", start=datetime.date(2011, 11, 1), end=datetime.date(2011, 11, 15)),
    Patch(name="Season 2", patch="1.0.0.129", start=datetime.date(2011, 11, 15), end=datetime.date(2011, 11, 29)),
    Patch(name="Season 2", patch="1.0.0.130", start=datetime.date(2011, 11, 29), end=datetime.date(2011, 12, 13)),
    Patch(name="Season 2", patch="1.0.0.131", start=datetime.date(2011, 12, 13), end=datetime.date(2012, 1, 17)),
    Patch(name="Season 2", patch="1.0.0.132", start=datetime.date(2012, 1, 17), end=datetime.date(2012, 2, 1)),
    Patch(name="Season 2", patch="1.0.0.133", start=datetime.date(2012, 2, 1), end=datetime.date(2012, 2, 14)),
    Patch(name="Season 2", patch="1.0.0.134", start=datetime.date(2012, 2, 14), end=datetime.date(2012, 2, 29)),
    Patch(name="Season 2", patch="1.0.0.135", start=datetime.date(2012, 2, 29), end=datetime.date(2012, 3, 20)),
    Patch(name="Season 2", patch="1.0.0.136", start=datetime.date(2012, 3, 20), end=datetime.date(2012, 4, 18)),
    Patch(name="Season 2", patch="1.0.0.138", start=datetime.date(2012, 4, 18), end=datetime.date(2012, 5, 1)),
    Patch(name="Season 2", patch="1.0.0.139", start=datetime.date(2012, 5, 1), end=datetime.date(2012, 5, 23)),
    Patch(name="Season 2", patch="1.0.0.140", start=datetime.date(2012, 5, 23), end=datetime.date(2012, 6, 5)),
    Patch(name="Season 2", patch="1.0.0.140b", start=datetime.date(2012, 6, 5), end=datetime.date(2012, 6, 17)),
    Patch(name="Season 2", patch="1.0.0.141", start=datetime.date(2012, 6, 17), end=datetime.date(2012, 7, 7)),
    Patch(name="Season 2", patch="1.0.0.142", start=datetime.date(2012, 7, 7), end=datetime.date(2012, 7, 19)),
    Patch(name="Season 2", patch="1.0.0.143", start=datetime.date(2012, 7, 19), end=datetime.date(2012, 8, 1)),
    Patch(name="Season 2", patch="1.0.0.144", start=datetime.date(2012, 8, 1), end=datetime.date(2012, 8, 14)),
    Patch(name="Season 2", patch="1.0.0.145", start=datetime.date(2012, 8, 14), end=datetime.date(2012, 8, 30)),
    Patch(name="Season 2", patch="1.0.0.146", start=datetime.date(2012, 8, 30), end=datetime.date(2012, 9, 12)),
    Patch(name="Season 2", patch="1.0.0.147", start=datetime.date(2012, 9, 12), end=datetime.date(2012, 9, 27)),
    Patch(name="Season 2", patch="1.0.0.148", start=datetime.date(2012, 9, 27), end=datetime.date(2012, 10, 17)),
    Patch(name="Season 2", patch="1.0.0.149", start=datetime.date(2012, 10, 17), end=datetime.date(2012, 10, 25)),
    Patch(name="Season 2", patch="1.0.0.150", start=datetime.date(2012, 10, 25), end=datetime.date(2012, 11, 13)),
    Patch(name="Season 2", patch="1.0.0.151", start=datetime.date(2012, 11, 13), end=datetime.date(2013, 2, 1)),
    Patch(name="Season 3", patch="3.01", start=datetime.date(2013, 2, 1), end=datetime.date(2013, 2, 13)),
    Patch(name="Season 3", patch="3.02", start=datetime.date(2013, 2, 13), end=datetime.date(2013, 3, 1)),
    Patch(name="Season 3", patch="3.03", start=datetime.date(2013, 3, 1), end=datetime.date(2013, 3, 19)),
    Patch(name="Season 3", patch="3.04", start=datetime.date(2013, 3, 19), end=datetime.date(2013, 3, 28)),
    Patch(name="Season 3", patch="3.5", start=datetime.date(2013, 3, 28), end=datetime.date(2013, 4, 30)),
    Patch(name="Season 3", patch="3.6", start=datetime.date(2013, 4, 30), end=datetime.date(2013, 5, 15)),
    Patch(name="Season 3", patch="3.7", start=datetime.date(2013, 5, 15), end=datetime.date(2013, 6, 13)),
    Patch(name="Season 3", patch="3.8", start=datetime.date(2013, 6, 13), end=datetime.date(2013, 7, 9)),
    Patch(name="Season 3", patch="3.9", start=datetime.date(2013, 7, 9), end=datetime.date(2013, 7, 30)),
    Patch(name="Season 3", patch="3.10", start=datetime.date(2013, 7, 30), end=datetime.date(2013, 8, 22)),
    Patch(name="Season 3", patch="3.10a", start=datetime.date(2013, 8, 22), end=datetime.date(2013, 9, 4)),
    Patch(name="Season 3", patch="3.11", start=datetime.date(2013, 9, 4), end=datetime.date(2013, 10, 1)),
    Patch(name="Season 3", patch="3.12", start=datetime.date(2013, 10, 1), end=datetime.date(2013, 10, 29)),
    Patch(name="Season 3", patch="3.13", start=datetime.date(2013, 10, 29), end=datetime.date(2014, 1, 15)),
    Patch(name="Season 4", patch="4.1", start=datetime.date(2014, 1, 15), end=datetime.date(2014, 2, 10)),
    Patch(name="Season 4", patch="4.2", start=datetime.date(2014, 2, 10), end=datetime.date(2014, 2, 27)),
    Patch(name="Season 4", patch="4.3", start=datetime.date(2014, 2, 27), end=datetime.date(2014, 3, 18)),
    Patch(name="Season 4", patch="4.4", start=datetime.date(2014, 3, 18), end=datetime.date(2014, 4, 3)),
    Patch(name="Season 4", patch="4.5", start=datetime.date(2014, 4, 3), end=datetime.date(2014, 4, 21)),
    Patch(name="Season 4", patch="4.6", start=datetime.date(2014, 4, 21), end=datetime.date(2014, 5, 8)),
    Patch(name="Season 4", patch="4.7", start=datetime.date(2014, 5, 8), end=datetime.date(2014, 5, 22)),
    Patch(name="Season 4", patch="4.8", start=datetime.date(2014, 5, 22), end=datetime.date(2014, 6, 4)),
    Patch(name="Season 4", patch="4.9", start=datetime.date(2014, 6, 4), end=datetime.date(2014, 6, 18)),
    Patch(name="Season 4", patch="4.10", start=datetime.date(2014, 6, 18), end=datetime.date(2014, 7, 2)),
    Patch(name="Season 4", patch="4.11", start=datetime.date(2014, 7, 2), end=datetime.date(2014, 7, 16)),
    Patch(name="Season 4", patch="4.12", start=datetime.date(2014, 7, 16), end=datetime.date(2014, 7, 30)),
    Patch(name="Season 4", patch="4.13", start=datetime.date(2014, 7, 30), end=datetime.date(2014, 8, 13)),
    Patch(name="Season 4", patch="4.14", start=datetime.date(2014, 8, 13), end=datetime.date(2014, 8, 27)),
    Patch(name="Season 4", patch="4.15", start=datetime.date(2014, 8, 27), end=datetime.date(2014, 9, 10)),
    Patch(name="Season 4", patch="4.16", start=datetime.date(2014, 9, 10), end=datetime.date(2014, 9, 25)),
    Patch(name="Season 4", patch="4.17", start=datetime.date(2014, 9, 25), end=datetime.date(2014, 10, 9)),
    Patch(name="Season 4", patch="4.18", start=datetime.date(2014, 10, 9), end=datetime.date(2014, 11, 5)),
    Patch(name="Season 4", patch="4.19", start=datetime.date(2014, 11, 5), end=datetime.date(2015, 1, 15)),
    Patch(name="Season 5", patch="5.1", start=datetime.date(2015, 1, 15), end=datetime.date(2015, 1, 28)),
    Patch(name="Season 5", patch="5.2", start=datetime.date(2015, 1, 28), end=datetime.date(2015, 2, 11)),
    Patch(name="Season 5", patch="5.3", start=datetime.date(2015, 2, 11), end=datetime.date(2015, 2, 25)),
    Patch(name="Season 5", patch="5.4", start=datetime.date(2015, 2, 25), end=datetime.date(2015, 3, 12)),
    Patch(name="Season 5", patch="5.5", start=datetime.date(2015, 3, 12), end=datetime.date(2015, 3, 25)),
    Patch(name="Season 5", patch="5.6", start=datetime.date(2015, 3, 25), end=datetime.date(2015, 4, 8)),
    Patch(name="Season 5", patch="5.7", start=datetime.date(2015, 4, 8), end=datetime.date(2015, 4, 28)),
    Patch(name="Season 5", patch="5.8", start=datetime.date(2015, 4, 28), end=datetime.date(2015, 5, 14)),
    Patch(name="Season 5", patch="5.9", start=datetime.date(2015, 5, 14), end=datetime.date(2015, 5, 28)),
    Patch(name="Season 5", patch="5.10", start=datetime.date(2015, 5, 28), end=datetime.date(2015, 6, 10)),
    Patch(name="Season 5", patch="5.11", start=datetime.date(2015, 6, 10), end=datetime.date(2015, 6, 24)),
    Patch(name="Season 5", patch="5.12", start=datetime.date(2015, 6, 24), end=datetime.date(2015, 7, 8)),
    Patch(name="Season 5", patch="5.13", start=datetime.date(2015, 7, 8), end=datetime.date(2015, 7, 22)),
    Patch(name="Season 5", patch="5.14", start=datetime.date(2015, 7, 22), end=datetime.date(2015, 8, 5)),
    Patch(name="Season 5", patch="5.15", start=datetime.date(2015, 8, 5), end=datetime.date(2015, 8, 20)),
    Patch(name="Season 5", patch="5.16", start=datetime.date(2015, 8, 20), end=datetime.date(2015, 9, 2)),
    Patch(name="Season 5", patch="5.17", start=datetime.date(2015, 9, 2), end=datetime.date(2015, 9, 16)),
    Patch(name="Season 5", patch="5.18", start=datetime.date(2015, 9, 16), end=datetime.date(2015, 9, 30)),
    Patch(name="Season 5", patch="5.19", start=datetime.date(2015, 9, 30), end=datetime.date(2015, 10, 14)),
    Patch(name="Season 5", patch="5.20", start=datetime.date(2015, 10, 14), end=datetime.date(2015, 10, 29)),
    Patch(name="Season 5", patch="5.21", start=datetime.date(2015, 10, 29), end=datetime.date(2016, 1, 14)),
    Patch(name="Season 6", patch="6.1", start=datetime.date(2016, 1, 14), end=datetime.date(2016, 1, 28)),
    Patch(name="Season 6", patch="6.2", start=datetime.date(2016, 1, 28), end=datetime.date(2016, 2, 10)),
    Patch(name="Season 6", patch="6.3", start=datetime.date(2016, 2, 10), end=datetime.date(2016, 2, 24)),
    Patch(name="Season 6", patch="6.4", start=datetime.date(2016, 2, 24), end=datetime.date(2016, 3, 9)),
    Patch(name="Season 6", patch="6.5", start=datetime.date(2016, 3, 9), end=datetime.date(2016, 3, 23)),
    Patch(name="Season 6", patch="6.6", start=datetime.date(2016, 3, 23), end=datetime.date(2016, 4, 6)),
    Patch(name="Season 6", patch="6.7", start=datetime.date(2016, 4, 6), end=datetime.date(2016, 4, 20)),
    Patch(name="Season 6", patch="6.8", start=datetime.date(2016, 4, 20), end=datetime.date(2016, 5, 4)),
    Patch(name="Season 6", patch="6.9", start=datetime.date(2016, 5, 4), end=datetime.date(2016, 5, 18)),
    Patch(name="Season 6", patch="6.10", start=datetime.date(2016, 5, 18), end=datetime.date(2016, 6, 1)),
    Patch(name="Season 6", patch="6.11", start=datetime.date(2016, 6, 1), end=datetime.date(2016, 6, 15)),
    Patch(name="Season 6", patch="6.12", start=datetime.date(2016, 6, 15), end=datetime.date(2016, 6, 29)),
    Patch(name="Season 6", patch="6.13", start=datetime.date(2016, 6, 29), end=datetime.date(2016, 7, 13)),
    Patch(name="Season 6", patch="6.14", start=datetime.date(2016, 7, 13), end=datetime.date(2016, 7, 26)),
    Patch(name="Season 6", patch="6.15", start=datetime.date(2016, 7, 26), end=datetime.date(2016, 8, 10)),
    Patch(name="Season 6", patch="6.16", start=datetime.date(2016, 8, 10), end=datetime.date(2016, 8, 24)),
    Patch(name="Season 6", patch="6.17", start=datetime.date(2016, 8, 24), end=datetime.date(2016, 9, 8)),
    Patch(name="Season 6", patch="6.18", start=datetime.date(2016, 9, 8), end=datetime.date(2016, 9, 21)),
    Patch(name="Season 6", patch="6.19", start=datetime.date(2016, 9, 21), end=datetime.date(2016, 10, 5)),
    Patch(name="Season 6", patch="6.20", start=datetime.date(2016, 10, 5), end=datetime.date(2016, 10, 19)),
    Patch(name="Season 6", patch="6.21", start=datetime.date(2016, 10, 19), end=datetime.date(2016, 11, 10)),
    Patch(name="Season 6", patch="6.22", start=datetime.date(2016, 11, 10), end=datetime.date(2016, 11, 22)),
    Patch(name="Season 6", patch="6.23", start=datetime.date(2016, 11, 22), end=datetime.date(2016, 12, 7)),
    Patch(name="Season 6", patch="6.24", start=datetime.date(2016, 12, 7), end=datetime.date(2017, 1, 11)),
    Patch(name="Season 7", patch="7.1", start=datetime.date(2017, 1, 11), end=datetime.date(2017, 1, 25)),
    Patch(name="Season 7", patch="7.2", start=datetime.date(2017, 1, 25), end=datetime.date(2017, 2, 8)),
    Patch(name="Season 7", patch="7.3", start=datetime.date(2017, 2, 8), end=datetime.date(2017, 2, 23)),
    Patch(name="Season 7", patch="7.4", start=datetime.date(2017, 2, 23), end=datetime.date(2017, 3, 8)),
    Patch(name="Season 7", patch="7.5", start=datetime.date(2017, 3, 8), end=datetime.date(2017, 3, 22)),
    Patch(name="Season 7", patch="7.6", start=datetime.date(2017, 3, 22), end=datetime.date(2017, 4, 5)),
    Patch(name="Season 7", patch="7.7", start=datetime.date(2017, 4, 5), end=datetime.date(2017, 4, 19)),
    Patch(name="Season 7", patch="7.8", start=datetime.date(2017, 4, 19), end=datetime.date(2017, 5, 3)),
    Patch(name="Season 7", patch="7.9", start=datetime.date(2017, 5, 3), end=datetime.date(2017, 5, 17)),
    Patch(name="Season 7", patch="7.10", start=datetime.date(2017, 5, 17), end=datetime.date(2017, 6, 1)),
    Patch(name="Season 7", patch="7.11", start=datetime.date(2017, 6, 1), end=datetime.date(2017, 6, 14)),
    Patch(name="Season 7", patch="7.12", start=datetime.date(2017, 6, 14), end=datetime.date(2017, 6, 28)),
    Patch(name="Season 7", patch="7.13", start=datetime.date(2017, 6, 28), end=datetime.date(2017, 7, 12)),
    Patch(name="Season 7", patch="7.14", start=datetime.date(2017, 7, 12), end=None)
]

Patch._Patch__patches = patches
